/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2005 NoMachine, http://www.nomachine.com.          */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef __Control_H__
#define __Control_H__

#include NXAGENT_NX_INCLUDE

#define MAX_CONNECTIONS 256

/*
 * Used to get the appropriate client to use as
 * resource in asynchronous "collect" operations
 * in nxcompext.
 */

#define nxagentDefaultClient() \
    (requestingClient != NULL ? requestingClient -> index : \
         serverClient -> index)

/*
 * Checked in the dispatcher to see if the client
 * has been put to sleep.
 */

#define nxagentSleepingClient(client) \
    (nxagentClientPriv((client)) -> is_ignored == True)

extern unsigned long nxagentStartTime;

extern Bool nxagentStrictControl;

extern unsigned int nxagentSyncCounter;
extern unsigned int nxagentSyncModulo;

extern long nxagentKarmaQueue[MAX_CONNECTIONS];

extern long nxagentStopKarmaSz;

extern Bool          nxagentUsesFocusKarma;
extern Bool          nxagentFocusKarmaOnStart;
extern Bool          nxagentFocusKarma;
extern unsigned long nxagentFocusKarmaTimeOut;

extern long nxagentStopKarmaSleep;
extern long nxagentStopKarmaSzDefault;
extern long nxagentStopKarmaSzCheck;

extern int nxagentClientPrivateIndex;

/*
 * Split images bigger than this size. The
 * client is set in the dispatcer, either
 * the main one or the one of MIT-SHM.
 */

extern int nxagentStopImgSz;

/*
 * Show an alert in the case a persistent
 * cache was not loaded by NX proxy at
 * startup.
 */

extern const char *nxagentNoCacheDialogText;
extern int         nxagentNoCacheDialogEnable;
extern int         nxagentNoCacheDialogTimeout;

extern void nxagentSetNoCacheDialogTimeout(int linkType);

extern void nxagentShowNoCacheDialog(void);

/*
 * Proxy sends a ClientMessage event to agent in case
 * it is not able to send more data on the proxy link.
 * Agent should stop accepting data from its clients
 * until the condition is ceased.
 */

extern Bool nxagentCongestion;

/*
 * Proxy informs agent when it is going through a reset.
 * This is intended to help agent to detect any X error
 * arising due to incomplete or completely lost requests.
 */

extern Bool nxagentReset;

enum ClientHint
{
  UNKNOWN = 0,
  OPEN_OFFICE,
  NXCLIENT,
  NXCLIENT_DIALOG
};

typedef struct{
  Bool is_ignored;
  pointer GetIFReply;
  enum ClientHint clientHint;
} nxagentPrivClient;

#define nxagentClientPriv(pClient) \
  ((nxagentPrivClient*)((pClient)->devPrivates[nxagentClientPrivateIndex].ptr))

#define nxagentWriteReplyToClient(pClient, size, pReply) { \
   if ((pClient)->swapped) \
      (*ReplySwapVector[X_GetInputFocus]) \
           (pClient, (int)(size), pReply); \
      else (void) WriteToClient(pClient, (int)(size), (char *)(pReply)); }

#define nxagentClientHint(pClient) \
    (nxagentClientPriv(pClient) -> clientHint)

#define nxagentUsesFrameBuffer(pClient) \
    (nxagentClientHint(pClient) == OPEN_OFFICE)

#define nxagentClientIsNXClientDialog(pClient) \
    (nxagentClientHint(pClient) == NXCLIENT_DIALOG)

void nxagentGuessClientHint(ClientPtr, Atom, char*);

/*
 * Update the proxy congestion flag by
 * checking if any congestion event is
 * become available.
 */

void nxagentCheckCongestion(void);

/*
 * Wakeup the sleeping clients.
 */

void nxagentWakeupByDecongestion(struct timeval **timeout);
void nxagentWakeupByReconnect(void);

/*
 * Set karma threshold according to
 * congestion state.
 */

void nxagentAdjustKarma(void);

/*
 * Suspend client or the entire agent.
 */

void nxagentCheckKarma(ClientPtr client, unsigned long now);

Bool nxagentSleepByBigReq(ClientPtr client);
Bool nxagentWakeByBigRequest(ClientPtr client);

Bool nxagentSleepByGetIFocus(ClientPtr client, pointer rep);
Bool nxagentWakeByGetIFocus(ClientPtr client);
int  nxagentProcGetInputFocus(ClientPtr client);

Bool nxagentSleepByKarma(ClientPtr client);
Bool nxagentWakeByKarma(ClientPtr client);
Bool nxagentWakeByError(ClientPtr client);

/*
 * Use this to terminate the agent
 * at the next dispatch loop.
 */

#define nxagentSetGiveUp() \
  do \
  { \
    dispatchException |= DE_TERMINATE; \
    isItTimeToYield = TRUE; \
  } \
  while (0)

/*
 * Used in PutImages bigger than MAX_REQUEST_SIZE(GCOps.c:Pixmap.c).
 */

#ifndef NXAGENT_MAX_REQUEST_EFF_SIZE
# define NXAGENT_MAX_REQUEST_EFF_SIZE \
    ( (MAX_REQUEST_SIZE << 2) - sizeof(xPutImageReq) )
#else
# error  NXAGENT_MAX_REQUEST_EFF_SIZE already defined !!
#endif

#ifndef MIN
# define MIN(A,B) \
    ( (A) < (B) ? (A) : (B) )
#endif

#endif /* __Control_H__ */
