/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2005 NoMachine, http://www.nomachine.com.          */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

/*

Copyright 1993 by Davor Matic

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation.  Davor Matic makes no representations about
the suitability of this software for any purpose.  It is provided "as
is" without express or implied warranty.

*/

#include <signal.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include <X11/X.h>
#include <X11/Xproto.h>
#include "screenint.h"
#include "input.h"
#include "misc.h"
#include "scrnintstr.h"
#include "servermd.h"

#ifdef WATCH
#include "unistd.h"
#endif

#include "NXalert.h"

#include "Agent.h"
#include "Display.h"
#include "Options.h"
#include "Error.h"
#include "Init.h"
#include "Args.h"
#include "Icon.h"
#include "Render.h"
#include "Font.h"
#include "Reconnect.h"
#include "Events.h"
#include "Dialog.h"
#include "Control.h"
#include "Splash.h"

#include NXAGENT_NX_INCLUDE
#include NXAGENT_NXLIB_INCLUDE

#include NXAGENT_ICON_NAME

#ifndef NXAGENT_NEVERSS
#include "screensaver"
#endif

/*
 * Set here the required log level.
 */

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG
#undef  WATCH

#ifdef NXAGENT_MVFB

Bool nxagentTrue24 = False;

#endif

Display *nxagentDisplay = NULL;
XVisualInfo *nxagentVisuals;
int nxagentNumVisuals;

static int nxagentXConnectionNumber;
static void nxagentAddXConnection(void);
void nxagentRemoveXConnection(void);
int nxagentIOErrorHandler(Display *display);

static Bool nxagentDisplayInfoSaved = False;
static Display *nxagentDisplayBackup = NULL;
static XlibGC nxagentBitmapGCBackup = NULL;
static XVisualInfo *nxagentVisualsRecBackup;
static int nxagentNumVisualsRecBackup;
static int nxagentNumDefaultColormapsRecBackup;
static int *nxagentDepthsRecBackup;
static int nxagentNumDepthsRecBackup;
static int nxagentNumPixmapFormatsRecBackup;
static XPixmapFormatValues *nxagentPixmapFormatsRecBackup;
static int DisplayWidthRecBackup;
static int DisplayHeightRecBackup;
static Bool nxagentRenderEnableRecBackup;
static Bool *nxagentVisualHasBeenIgnored;

static enum
{
  NOTHING = 0,
  OPENED,
  GOT_VISUAL_INFO,
  ALLOC_DEF_COLORMAP,
  GOT_DEPTH_LIST,
  GOT_PIXMAP_FORMAT_LIST,
  EVERYTHING_DONE
} reconnectDisplayState;

int nxagentDefaultVisualIndex;
Colormap *nxagentDefaultColormaps;
int nxagentNumDefaultColormaps;
int *nxagentDepths;
int nxagentNumDepths;
XPixmapFormatValues *nxagentPixmapFormats;
int nxagentNumPixmapFormats;
Pixel nxagentBlackPixel;
Pixel nxagentWhitePixel;
Drawable nxagentDefaultDrawables[MAXDEPTH + 1];
Pixmap nxagentScreenSaverPixmap;

/*
 * Also used in Cursor.c. There are huge problems
 * using GC definition. This is to be reworked.
 */

XlibGC nxagentBitmapGC;

/*
 * The "confine" window is used in the nxagentConstrainCursor
 * procedure. We are currently overriding the original Xnest
 * behaviour. It is unclear what this window is used for.
 */

Window nxagentConfineWindow;

Pixmap nxagentIconPixmap;
Pixmap nxagentIconShape;
Bool useXpmIcon = False;

unsigned int nxagentLogoColor(unsigned int value);
Bool nxagentMakeIcon(Display *display, Pixmap *nxIcon, Pixmap *nxMask);

/*
 * FIXME: These must definitely become local.
 */

XVisualInfo pV;
unsigned int r, b, g, or, ob, og, off;

/*
 * Other local functions.
 */

Display *nxagentSafeOpenDisplay(char *display);
void nxagentRaiseSignal(int signal);
void nxagentDisconnectDisplay(void);
Bool nxagentReconnectDisplay(void *p0);

/*
 * This is located in connection.c.
 */

extern void RejectWellKnownSockets(void);

#ifdef NXAGENT_TIMESTAMP
unsigned long startTime;
#endif

unsigned int nxagentLogoColor(unsigned int value)
{
  /*
   * Takes a color value in RGB24 (0xff0000, 0x00ff00,
   * 0x0000ff) and converts it into an equivalent for
   * the current visual.
   */

  int cr=0, cg=0, cb=0;

  cr = (value >> or) &r;
  cg = (value >> (og - 8)) &g;
  cb = (value >> (ob - 16)) &b;

  return (cr | cg | cb);
}

/*
 * FIXME: This must be moved to Error.c.
 */

int nxagentIOErrorHandler(Display *display)
{
  if (nxagentOption(Persistent) && nxagentSessionState != SESSION_STARTING)
  {
    if (nxagentSessionState == SESSION_UP)
    {
      fprintf(stderr, "Info: IOError while processing events or requests.\n");
      fprintf(stderr, "Info: Suspending session on IOError.\n");

      nxagentDisconnect();
    }
    else if (nxagentSessionState == SESSION_GOING_DOWN)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentIOErrorHandler: Got IOError session state is [SESSION_GOING_DOWN].\n");
      #endif
    }
    else if (nxagentSessionState == SESSION_GOING_UP)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentIOErrorHandler: Got IOError session state is [SESSION_GOING_UP].\n");
      #endif

      nxagentSessionState = SESSION_GOING_DOWN;

      nxagentSetReconnectError(FAILED_RESUME_DISPLAY_BROKEN_ALERT, "Got IO Error during reconnect.");

      nxagentChangeOption(Fullscreen, False);

      return 1;
    }
    else if (nxagentSessionState == SESSION_DOWN)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentIOErrorHandler: Got IOError session state is [SESSION_DOWN] - Ignoring it.\n");
      #endif

      return 1;
    }
    else
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentIOErrorHandler: Got IOError session state is %d.\n", nxagentSessionState);
      #endif
    }

    nxagentRemoveXConnection();
    nxagentSessionState = SESSION_DOWN;

    fprintf(stderr, "Info: Session suspended.\n");

    return 1;
  }

  XSetIOErrorHandler(NULL);

  CloseWellKnownConnections();

  FatalError("NXAGENT: Fatal IO error on display \"%s\".\n", nxagentDisplayName);

  return -1;
}

/*
 * Force a shutdown of any connection attempt
 * while connecting to the remote display.
 * This is needed to avoid a hang up in case
 * of loopback connections to our own listen-
 * ing sockets.
 */

static void nxagentRejectConnection(int signal)
{
  #ifdef TEST
  fprintf(stderr, "nxagentRejectConnection: Going to reject client connections.\n");
  #endif

  #ifndef NXAGENT_UPGRADE
  /*
   * FIXME: We haven't finised to
   *        port the os directory
   *        to the new Xorg tree.
   *        This function is still
   *        not available in the new
   *        tree.
   */

  RejectWellKnownSockets();
  #endif

  #ifdef TEST
  fprintf(stderr, "nxagentRejectConnection: Setting new alarm to 5 seconds from now.\n");
  #endif

  /*
   * A further timeout is unlikely to happen
   * in the case of loopback connections.
   */

  alarm(5);
}

static void nxagentSigchldHandler(int signal)
{
  int pid = 0;
  int status;
  int options;

  #ifdef TEST
  fprintf(stderr, "nxagentSigchldHandler: Going to check the children processes.\n");
  #endif

  options = WNOHANG | WUNTRACED;

  /*
   * Try with the pid of the dialog process.
   * Leave the other children unaffected.
   */

  if (pid == 0 && nxagentRootlessDialogPid)
  {
    pid = waitpid(nxagentRootlessDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Rootless dialog).\n",
                  nxagentRootlessDialogPid);
      #endif

      pid = nxagentRootlessDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentPulldownDialogPid)
  {
    pid = waitpid(nxagentPulldownDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Pulldown dialog).\n",
                  nxagentPulldownDialogPid);
      #endif

      pid = nxagentPulldownDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentKillDialogPid)
  {
    pid = waitpid(nxagentKillDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Kill dialog).\n",
                  nxagentKillDialogPid);
      #endif

      pid = nxagentKillDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentSuspDialogPid)
  {
    pid = waitpid(nxagentSuspDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (Suspend dialog).\n",
                  nxagentSuspDialogPid);
      #endif

      pid = nxagentSuspDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentSlowModeDialogPid)
  {
    pid = waitpid(nxagentSlowModeDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (SlowMode dialog).\n",
                  nxagentSlowModeDialogPid);
      #endif

      pid = nxagentSlowModeDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentFastModeDialogPid)
  {
    pid = waitpid(nxagentFastModeDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (FastMode dialog).\n",
                  nxagentFastModeDialogPid);
      #endif

      pid = nxagentFastModeDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentEnableRandRModeDialogPid)
  {
    pid = waitpid(nxagentEnableRandRModeDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (EnableRandRMode dialog).\n",
                  nxagentEnableRandRModeDialogPid);
      #endif

      pid = nxagentEnableRandRModeDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentDisableRandRModeDialogPid)
  {
    pid = waitpid(nxagentDisableRandRModeDialogPid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (DisableRandRMode dialog).\n",
                  nxagentDisableRandRModeDialogPid);
      #endif

      pid = nxagentDisableRandRModeDialogPid = 0;
    }
  }

  if (pid == 0 && nxagentEnableLazyModePid)
  {
    pid = waitpid(nxagentEnableLazyModePid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (EnableLazyMode dialog).\n",
                  nxagentEnableLazyModePid);
      #endif

      pid = nxagentEnableLazyModePid = 0;
    }
  }

  if (pid == 0 && nxagentDisableLazyModePid)
  {
    pid = waitpid(nxagentDisableLazyModePid, &status, options);

    if (pid == -1 && errno == ECHILD)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Got ECHILD waiting for child %d (DisableLazyMode dialog).\n",
                  nxagentDisableLazyModePid);
      #endif

      pid = nxagentDisableLazyModePid = 0;
    }
  }

  if (pid == -1)
  {
    FatalError("nxagentSigchldHandler: Got error %s waiting for child.\n", strerror(errno));
  }

  if (pid > 0)
  {
    if (WIFSTOPPED(status))
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentSigchldHandler: Child process [%d] was stopped "
                  "with signal [%d].\n", pid, (WSTOPSIG(status)));
      #endif
    }
    else
    {
      #ifdef TEST

      if (WIFEXITED(status))
      {
        fprintf(stderr, "nxagentSigchldHandler: Child process [%d] exited "
                    "with status [%d].\n", pid, (WEXITSTATUS(status)));
      }
      else if (WIFSIGNALED(status))
      {
        fprintf(stderr, "nxagentSigchldHandler: Child process [%d] died "
                    "because of signal [%d].\n", pid, (WTERMSIG(status)));
      }

      #endif

      nxagentResetDialog(pid);
    }
  }
  #ifdef TEST
  else if (pid == 0)
  {
    fprintf(stderr, "nxagentSigchldHandler: No own child process exited.\n");
  }
  #endif

  return;
}

Display *nxagentSafeOpenDisplay(char *display)
{
  Display *ret;
  struct sigaction sa_new, sa_old;
  int res;

  sa_new.sa_handler = nxagentRejectConnection;
  sigfillset(&sa_new.sa_mask);
  sa_new.sa_flags = SA_RESTART;

  while (((res = sigaction(SIGALRM, &sa_new, &sa_old)) == -1) &&
             (errno == EINTR));

  if (res == -1)
  {
    FatalError("Error: Can't set alarm for rejecting connections.");
  }

  alarm(10);

  ret = XOpenDisplay(display);

  alarm(0);

  while (((res = sigaction(SIGALRM, &sa_old, NULL)) == -1) &&
             (errno == EINTR));

  if (res == -1)
  {
    FatalError("Error: Can't reset alarm for rejecting connections.");
  }

  /*
   * We want the Xlib I/O error handler to return,
   * instead of quitting the application. The other
   * alternative is to use setjmp()/longjmp() but
   * this leaves the door open to unexpected bugs
   * when dealing with interaction with other X
   * server layers.
   */

  NXContinueOnDisplayError(1);

  /*
   * By default nxcomp installs its signal handlers.
   * We let the default X server's handler manage
   * the SIGINT, SIGTERM and SIGHUP.
   *
   * The following functions will either save the old
   * handlers or, alternatively, won't do anything
   * if the NX transport isn't running, so the agent
   * should create its handlers before calling them.
   */

  NXTransSignal(SIGINT,  NX_SIGNAL_DISABLE);
  NXTransSignal(SIGTERM, NX_SIGNAL_DISABLE);
  NXTransSignal(SIGHUP,  NX_SIGNAL_DISABLE);

  /*
   * The agent needs also to be notified about its
   * children, so that it can waitpid() for them.
   */

  sa_new.sa_handler = nxagentSigchldHandler;

  sigfillset(&sa_new.sa_mask);
  sa_new.sa_flags = SA_RESTART;

  while (((res = sigaction(SIGCHLD, &sa_new, NULL)) == -1) &&
             (errno == EINTR));

  if (res == -1)
  {
    FatalError("Error: Can't set signal handler for children.");
  }

  NXTransSignal(SIGCHLD, NX_SIGNAL_FORWARD);

  #ifdef TEST
  fprintf(stderr, "nxagentSafeOpenDisplay: Successfully opened display at [%p].\n",
              (void *) ret);
  #endif

  return ret;
}

void nxagentRaiseSignal(int signal)
{
  if (kill(getpid(), signal) == -1)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentRaiseSignal: Couldn't raise signal %d: %s.\n",
                signal, strerror(errno));
    #endif
  }
}

void nxagentOpenDisplay(int argc, char *argv[])
{
  XVisualInfo vi;
  long mask;
  int i;

  if (!nxagentDoFullGeneration) return;

#ifdef NXAGENT_TIMESTAMP
  startTime = GetTimeInMillis();
  fprintf(stderr, "Display: Going to open Display on real X server time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif

  /*
   * Initialize the SIGHUP handler for both the
   * cases: Pesistent and not Persistent sessions.
   */

  nxagentInitSighupHandler();

  /*
   * Initialize the Reconnector
   * only for Persistent sessions.
   */

  if (nxagentOption(Persistent))
  {
    nxagentInitReconnector();
  }

  if (*nxagentDisplayName == '\0')
  {
    strncpy(nxagentDisplayName, XDisplayName(NULL), 1023);

    nxagentDisplayName[1023] = '\0';
  }

  XSetIOErrorHandler(nxagentIOErrorHandler);

  nxagentCloseDisplay();

  nxagentDisplay = nxagentSafeOpenDisplay(nxagentDisplayName);

  if (nxagentDisplay == NULL)
  {
    FatalError("NXAGENT: Unable to open display \"%s\".\n", nxagentDisplayName);
  }

  nxagentXConnectionNumber = XConnectionNumber(nxagentDisplay);

  #ifdef TEST
  fprintf(stderr, "nxagentOpenDisplay: Display image order is [%d] bitmap order is [%d].\n",
              ImageByteOrder(nxagentDisplay), BitmapBitOrder(nxagentDisplay));

  fprintf(stderr, "nxagentOpenDisplay: Display scanline unit is [%d] scanline pad is [%d].\n",
              BitmapUnit(nxagentDisplay), BitmapPad(nxagentDisplay));
  #endif

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 1.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
#1   U  1	1	256 bits (0 KB) ->	150 bits (0 KB) ->	256/1 -> 150/1	= 1.707:1
#20     1	1	119104 bits (15 KB) ->	28 bits (0 KB) ->	119104/1 -> 28/1	= 4253.714:1
#98     2		512 bits (0 KB) ->	84 bits (0 KB) ->	256/1 -> 42/1	= 6.095:1
*/

  sleep(60);

  #endif

#ifdef NXAGENT_ERRORH
  XSetErrorHandler((XErrorHandler)nxagentErrorHandler);
#endif

#ifdef NXAGENT_TIMESTAMP
  fprintf(stderr, "Display: display on real X opened, time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif

  nxagentUseNXTrans =
      nxagentPostProcessArgs(nxagentDisplayName, nxagentDisplay,
                                 DefaultScreenOfDisplay(nxagentDisplay));

  if (nxagentSynchronize)
    XSynchronize(nxagentDisplay, True);

  mask = VisualScreenMask;
  vi.screen = DefaultScreen(nxagentDisplay);
  nxagentVisuals = XGetVisualInfo(nxagentDisplay, mask, &vi, &nxagentNumVisuals);
  if (nxagentNumVisuals == 0 || nxagentVisuals == NULL)
    FatalError("Unable to find any visuals.\n");

  if (nxagentUserDefaultClass || nxagentUserDefaultDepth) {
    nxagentDefaultVisualIndex = UNDEFINED;
    for (i = 0; i < nxagentNumVisuals; i++)
      if ((!nxagentUserDefaultClass ||
           nxagentVisuals[i].class == nxagentDefaultClass)
          &&
          (!nxagentUserDefaultDepth ||
           nxagentVisuals[i].depth == nxagentDefaultDepth)) {
        nxagentDefaultVisualIndex = i;
        break;
      }
    if (nxagentDefaultVisualIndex == UNDEFINED)
      FatalError("Unable to find desired default visual.\n");
  }
  else {
    vi.visualid = XVisualIDFromVisual(DefaultVisual(nxagentDisplay,
                                      DefaultScreen(nxagentDisplay)));
    nxagentDefaultVisualIndex = 0;
    for (i = 0; i < nxagentNumVisuals; i++)
      if (vi.visualid == nxagentVisuals[i].visualid)
        nxagentDefaultVisualIndex = i;
  }

  nxagentNumDefaultColormaps = nxagentNumVisuals;
  nxagentDefaultColormaps = (Colormap *)xalloc(nxagentNumDefaultColormaps *
                                             sizeof(Colormap));
  for (i = 0; i < nxagentNumDefaultColormaps; i++)
    nxagentDefaultColormaps[i] = XCreateColormap(nxagentDisplay,
                                               DefaultRootWindow(nxagentDisplay),
                                               nxagentVisuals[i].visual,
                                               AllocNone);

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 4.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

  nxagentBlackPixel = BlackPixel(nxagentDisplay, DefaultScreen(nxagentDisplay));
  nxagentWhitePixel = WhitePixel(nxagentDisplay, DefaultScreen(nxagentDisplay));

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 5.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

  /*
   * Initialize the agent's event mask that will be requested
   * for the root and all the top level windows. If the nested
   * window is a child of an existing window, we will need to
   * receive StructureNotify events. If we are going to manage
   * the changes in root window's visibility we'll also need
   * VisibilityChange events.
   */

/*
FIXME: Use of nxagentParentWindow is strongly deprecated.
       We need also to clarify which events are selected
       in the diferent operating modes.
*/

  nxagentInitDefaultEventMask();

  /*
   * Get the pixmap depths and format supported
   * by the real display.
   */

  nxagentGetDepthsAndPixmapFormats();

  /*
   * Create a pixmap for each depth matching the
   * local supported formats with format available
   * on the remote display.
   */

  nxagentSetDefaultDrawables();

  #ifdef RENDER
  if (nxagentRenderEnable)
  {
    nxagentRenderExtensionInit();
  }
  #endif

  /*
   * This GC is referenced in Cursor.c. It can be
   * probably removed.
   */

  nxagentBitmapGC = XCreateGC(nxagentDisplay, nxagentDefaultDrawables[1], 0L, NULL);

  /*
   * Note that this "confine window" is useless at the
   * moment as we reimplement nxagentConstrainCursor()
   * to skip the "constrain" stuff.
   */

  #ifdef TEST
  fprintf(stderr, "nxagentOpenDisplay: Going to create agent's confine window.\n");
  #endif

  nxagentConfineWindow = XCreateWindow(nxagentDisplay,
                                       DefaultRootWindow(nxagentDisplay),
                                       0, 0, 1, 1, 0, 0,
                                       InputOnly,
                                       CopyFromParent,
                                       0L, NULL);

  #ifdef TEST
  fprintf(stderr, "nxagentOpenDisplay: Created agent's confine window with id [%ld].\n",
              nxagentConfineWindow);
  #endif

  if (!(nxagentUserGeometry.flag & XValue))
  {
    nxagentChangeOption(RootX, 0);
  }

  if (!(nxagentUserGeometry.flag & YValue))
  {
    nxagentChangeOption(RootY, 0);
  }

  if (nxagentParentWindow == 0)
  {
    if (!(nxagentUserGeometry.flag & WidthValue))
    {
      if (nxagentOption(Fullscreen))
      {
        nxagentChangeOption(RootWidth, DisplayWidth(nxagentDisplay, DefaultScreen(nxagentDisplay)));
      }
      else
      {
        nxagentChangeOption(RootWidth, 3 * DisplayWidth(nxagentDisplay,
                                           DefaultScreen(nxagentDisplay)) / 4);
      }
    }

    if (!(nxagentUserGeometry.flag & HeightValue))
    {
      if (nxagentOption(Fullscreen))
      {
        nxagentChangeOption(RootHeight, DisplayHeight(nxagentDisplay, DefaultScreen(nxagentDisplay)));
      }
      else
      {
        nxagentChangeOption(RootHeight, 3 * DisplayHeight(nxagentDisplay,
                                            DefaultScreen(nxagentDisplay)) / 4);
      }
    }
  }

  if (!nxagentUserBorderWidth)
  {
    nxagentChangeOption(BorderWidth, 1);
  }

  nxagentLogoDepth = DefaultDepth(nxagentDisplay,
                               DefaultScreen(nxagentDisplay)
                              );

  pV = nxagentVisuals[nxagentDefaultVisualIndex];

  r = pV.red_mask;
  g = pV.green_mask;
  b = pV.blue_mask;
  if(!pV.red_mask|| !pV.green_mask|| !pV.blue_mask){
      nxagentLogoBlack = 0x000000;
      nxagentLogoRed   = 0xff0000;
      nxagentLogoWhite = 0xffffff;
  }
  else
  {
      for(or=0,off=0x800000;(r&(off>>or)) == 0; or++);
      for(og=0,off=0x800000;(g&(off>>og)) == 0; og++);
      for(ob=0,off=0x800000;(b&(off>>ob)) == 0; ob++);
      nxagentLogoRed   = nxagentLogoColor(0xff0000);
      nxagentLogoBlack = nxagentLogoColor(0x000000);
      nxagentLogoWhite = 0xffffff;
  }

#ifdef NXAGENT_LOGO_DEBUG
  fprintf(stderr, "mask r: %x g: %x b: %x\n", r, g, b);
  fprintf(stderr, "offset r: %d g: %d b: %d\n", or, og, ob);
#endif

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 5.1.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

  useXpmIcon = nxagentMakeIcon(nxagentDisplay, &nxagentIconPixmap, &nxagentIconShape);

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 5.2.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
#84     2		512 bits (0 KB) ->	76 bits (0 KB) ->	256/1 -> 38/1	= 6.737:1
*/

  sleep(30);

  #endif

#ifndef NXAGENT_NEVERSS
  nxagentScreenSaverPixmap =
    XCreatePixmapFromBitmapData(nxagentDisplay,
                                DefaultRootWindow(nxagentDisplay),
                                (char *)screensaver_bits,
                                screensaver_width,
                                screensaver_height,
                                nxagentWhitePixel,
                                nxagentBlackPixel,
                                DefaultDepth(nxagentDisplay,
                                DefaultScreen(nxagentDisplay)));
#endif

  #ifdef WATCH

  fprintf(stderr, "nxagentOpenDisplay: Watchpoint 6.\n");

/*
Reply   Total	Cached	Bits In			Bits Out		Bits/Reply	  Ratio
------- -----	------	-------			--------		----------	  -----
N/A
*/

  sleep(30);

  #endif

#ifdef NXAGENT_TIMESTAMP
  fprintf(stderr, "Display: nxagentOpenDisplay finished, time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif
  if (nxagentOption(Persistent))
    reconnectDisplayState = EVERYTHING_DONE;
}

void nxagentGetDepthsAndPixmapFormats()
{
  nxagentDepths = XListDepths(nxagentDisplay, DefaultScreen(nxagentDisplay),
                                  &nxagentNumDepths);

  nxagentPixmapFormats = XListPixmapFormats(nxagentDisplay, &nxagentNumPixmapFormats);

  if (nxagentDepths == NULL || nxagentPixmapFormats == NULL)
  {
    #ifdef PANIC
    fprintf(stderr, "nxagentGetDepthsAndPixmapFormats: PANIC! Failed to get available depths and pixmap formats.\n");
    #endif

    FatalError("Error: Failed to get available depths and pixmap formats.");
  }

  #ifdef TEST
  fprintf(stderr, "nxagentGetDepthsAndPixmapFormats: Got [%d] depths and [%d] available pixmap formats.\n",
              nxagentNumDepths, nxagentNumPixmapFormats);
  #endif
}

void nxagentSetDefaultDrawables()
{
  int i, j;

  for (i = 0; i <= MAXDEPTH; i++)
  {
    nxagentDefaultDrawables[i] = None;
  }

  for (i = 0; i < nxagentNumPixmapFormats; i++)
  {
    #ifdef TEST
    fprintf(stderr, "nxagentSetDefaultDrawables: Checking remote pixmap format [%d] with depth [%d] "
                "bits per pixel [%d] scanline pad [%d].\n", i, nxagentPixmapFormats[i].depth,
                    nxagentPixmapFormats[i].bits_per_pixel, nxagentPixmapFormats[i].scanline_pad);
    #endif

    if (nxagentPixmapFormats[i].depth == 24)
    {
      if (nxagentPixmapFormats[i].bits_per_pixel == 24)
      {
        #ifdef TEST
        fprintf(stderr, "nxagentSetDefaultDrawables: WARNING! Assuming remote pixmap "
                    "format [%d] as true 24 bits.\n", i);
        #endif

        nxagentTrue24 = True;
      }
    }

    for (j = 0; j < nxagentNumDepths; j++)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSetDefaultDrawables: Cheching depth at index [%d] with pixmap depth [%d] "
                  "and display depth [%d].\n", j, nxagentPixmapFormats[i].depth, nxagentDepths[j]);
      #endif

      if ((nxagentPixmapFormats[i].depth == 1 ||
              nxagentPixmapFormats[i].depth == nxagentDepths[j]) &&
                  nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] == None)
      {
        nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] =
            XCreatePixmap(nxagentDisplay, DefaultRootWindow(nxagentDisplay),
                              1, 1, nxagentPixmapFormats[i].depth);

        #ifdef TEST
        fprintf(stderr, "nxagentSetDefaultDrawables: Created default drawable for depth [%d].\n",
                    nxagentPixmapFormats[i].depth);
        #endif
      }
    }

    if (nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] == None)
    {
      #ifdef TEST
      fprintf(stderr, "nxagentSetDefaultDrawables: WARNING! Forcing default drawable for depth [%d].\n",
                  nxagentPixmapFormats[i].depth);
      #endif

      nxagentDefaultDrawables[nxagentPixmapFormats[i].depth] =
          XCreatePixmap(nxagentDisplay, DefaultRootWindow(nxagentDisplay),
                            1, 1, nxagentPixmapFormats[i].depth);
    }
  }
}

void nxagentCloseDisplay()
{
  #ifdef TEST
  fprintf(stderr, "nxagentCloseDisplay: Called with full generation [%d] and display [%p].\n",
              nxagentDoFullGeneration, (void *) nxagentDisplay);
  #endif

  if (!nxagentDoFullGeneration || !nxagentDisplay) return;

  /*
    If nxagentDoFullGeneration all x resources will be destroyed upon closing
    the display connection.  There is no need to generate extra protocol.
    */

/*
FIXME: These make nxagent go in segfault on exit.
       Reason must be investigated.

  xfree(nxagentDefaultColormaps);
  XFree(nxagentVisuals);
*/
  XFree(nxagentDepths);
  XFree(nxagentPixmapFormats);
  nxagentFreeFontCache();

/*
  nxagentFreeFontMatchStuff();
*/
  XCloseDisplay(nxagentDisplay);

  /*
   * Wait for the NX transport
   * to shut down, if any.
   */

  #ifdef TEST
  fprintf(stderr, "nxagentCloseDisplay: Waiting for the NX transport to terminate.\n");
  #endif

  NXTransDestroy();

  #ifdef TEST
  fprintf(stderr, "nxagentCloseDisplay: Setting the display to NULL.\n");
  #endif

  nxagentDisplay = NULL;
}

Bool nxagentMakeIcon(Display *display, Pixmap *nxIcon, Pixmap *nxMask)
{
  char *env_path = getenv("PATH");
  int lenght_env_path = 0;
  char icon_filename [256];
  char default_path [256];
  char *icon_path = malloc( strlen(env_path) + sizeof(icon_filename) );
  FILE *icon_fp;
  int status;
  Bool existXpmIcon = False;

  Pixmap IconPixmap;
  Pixmap IconShape;

  if (env_path == NULL)
    lenght_env_path = 0;
  else
    lenght_env_path = strlen(env_path) + 1;
  strncpy(icon_filename, "", 255);
  strncpy(default_path, "", 255);

  strcat(icon_filename, NXAGENT_ICON_NAME);
  strcat(default_path,"/usr/NX/share/images/");
  strcat(default_path,icon_filename);

  if ((icon_fp = fopen(default_path, "r")) == NULL)
  {
    char *s;
    char *temp_path = malloc(lenght_env_path + strlen(icon_filename) );
    char *temp_path1 = malloc(lenght_env_path + strlen(icon_filename) );

    strncpy(temp_path, env_path, strlen(env_path));
    strncpy(temp_path1, "", lenght_env_path + strlen(icon_filename) );

    while (strlen(temp_path) > 0)
    {
       s = strpbrk (temp_path, ":");
       if (s == NULL) break;

       strncpy (temp_path1, temp_path , strlen(temp_path) - strlen(s) );
       strncat (temp_path1, "/", 1);
       strncat (temp_path1, icon_filename, strlen(icon_filename));
       if ((icon_fp = fopen(temp_path1, "r")) != NULL)
       {
          fclose (icon_fp);
          existXpmIcon = True;
          strcpy(icon_path,temp_path1);
          break;
       }
       strncpy(temp_path1, "", lenght_env_path + strlen(icon_filename) );
       strncpy(temp_path1, s + 1, strlen(s)-1);
       strncpy(temp_path, "", lenght_env_path + strlen(icon_filename) );
       strcpy(temp_path, temp_path1 );
       strncpy(temp_path1, "", lenght_env_path + strlen(icon_filename) );
     }
     free(temp_path);
     free(temp_path1);
  }
  else
  {
     fclose (icon_fp);
     existXpmIcon = True;
     strcpy(icon_path, default_path);
  }

  if (existXpmIcon)
  {
     status = XpmReadFileToPixmap(display,
                                 DefaultRootWindow(display),
                               icon_path,
                               &IconPixmap,
                               &IconShape,
                               NULL);

     if (status != XpmSuccess) {
        fprintf(stderr, "XpmError: %s\n", XpmGetErrorString(status));
        existXpmIcon = False;
     }
  }

  if (!existXpmIcon)
  {
     status = XpmCreatePixmapFromData(display,
                                        DefaultRootWindow(display),
                                        nxagentIconData,
                                        &IconPixmap,
                                        &IconShape,
                                        NULL);

     if (status != XpmSuccess)
     {
        fprintf(stderr, "XpmError: %s\n", XpmGetErrorString(status));

        existXpmIcon = False;
     }
     else
     {
        existXpmIcon = True;
     }
  }

  free(icon_path);
  *nxIcon = IconPixmap;
  *nxMask = IconShape;
  return existXpmIcon;
}

Bool nxagentXServerGeometryChanged()
{
  return  (WidthOfScreen(DefaultScreenOfDisplay(nxagentDisplay)) !=
               WidthOfScreen(DefaultScreenOfDisplay(nxagentDisplayBackup))) ||
                    (HeightOfScreen(DefaultScreenOfDisplay(nxagentDisplay)) !=
                         HeightOfScreen(DefaultScreenOfDisplay(nxagentDisplayBackup)));
}

void nxagentBackupDisplayInfo(void)
{
  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DEBUG)
  fprintf(stderr, "nxagentBackupDisplayInfo\n");
  #endif

  if (nxagentDisplayInfoSaved)
  {
    return;
  }

  /*
   * Since we need the display structure
   * in order to behave correctly when no X
   * connection is available, we must always
   * have a good display record.
   * It can be discarded only when a new X
   * connection is available, so we store it
   * in order to destroy whenever the recon-
   * nection succed.
   */

  nxagentDisplayBackup = nxagentDisplay;
  nxagentBitmapGCBackup = nxagentBitmapGC;
  nxagentPixmapFormatsRecBackup = nxagentPixmapFormats;
  nxagentNumPixmapFormatsRecBackup = nxagentNumPixmapFormats;
  nxagentDepthsRecBackup = nxagentDepths;
  nxagentNumDepthsRecBackup = nxagentNumDepths;
  nxagentNumDefaultColormapsRecBackup = nxagentNumDefaultColormaps;
  nxagentVisualsRecBackup = nxagentVisuals;
  nxagentNumVisualsRecBackup = nxagentNumVisuals;
  if (nxagentVisualHasBeenIgnored)
  {
    xfree(nxagentVisualHasBeenIgnored);
    nxagentVisualHasBeenIgnored = NULL;
  }
  nxagentVisualHasBeenIgnored = xalloc(nxagentNumVisuals * sizeof(Bool));
  DisplayWidthRecBackup = DisplayWidth(nxagentDisplay, DefaultScreen(nxagentDisplay));
  DisplayHeightRecBackup = DisplayHeight(nxagentDisplay, DefaultScreen(nxagentDisplay));
  nxagentRenderEnableRecBackup = nxagentRenderEnable;

  nxagentDisplayInfoSaved = True;
}

void nxagentCleanupBackupDisplayInfo(void)
{
  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DISPLAY_DEBUG)
  fprintf(stderr, "nxagentCleanupBackupDisplayInfo\n");
  #endif

  xfree(nxagentPixmapFormatsRecBackup);
  nxagentNumPixmapFormatsRecBackup = 0;

  xfree(nxagentDepthsRecBackup);
  nxagentNumDepthsRecBackup = 0;

  nxagentNumDefaultColormapsRecBackup = 0;

  xfree(nxagentVisualsRecBackup);
  nxagentNumVisualsRecBackup = 0;

  if (nxagentVisualHasBeenIgnored)
  {
    xfree(nxagentVisualHasBeenIgnored);
    nxagentVisualHasBeenIgnored = NULL;
  }

  DisplayWidthRecBackup = 0;
  DisplayHeightRecBackup = 0;

  if (nxagentDisplayBackup)
  {
    XCloseDisplay(nxagentDisplayBackup);

    nxagentDisplayBackup = NULL;
  }

  if (nxagentBitmapGCBackup)
  {
    if (nxagentDisplayBackup)
    {
      XFreeGC(nxagentDisplayBackup, nxagentBitmapGCBackup);
    }
    else
    {
      xfree(nxagentBitmapGCBackup);
    }

    nxagentBitmapGCBackup = NULL;
  }

  nxagentDisplayInfoSaved = False;
}

void nxagentDisconnectDisplay(void)
{
  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DISPLAY_DEBUG)
  fprintf(stderr, "nxagentDisconnectDisplay\n");
  #endif

  switch (reconnectDisplayState)
  {
    case EVERYTHING_DONE:
      if (nxagentBitmapGC &&
              nxagentBitmapGCBackup &&
                  (nxagentBitmapGC != nxagentBitmapGCBackup))
      {
        XFreeGC(nxagentDisplay, nxagentBitmapGC);
      }

      nxagentBitmapGC = nxagentBitmapGCBackup;
    case GOT_PIXMAP_FORMAT_LIST:
    case GOT_DEPTH_LIST:
    case ALLOC_DEF_COLORMAP:
      if (nxagentDefaultColormaps)
      {
        int i;

        for (i = 0; i < nxagentNumDefaultColormaps; i++)
        {
          nxagentDefaultColormaps[i] = None;
        }
      }
    case GOT_VISUAL_INFO:
    case OPENED:
      /*
       * Actually we need the nxagentDisplay structure
       * in order to let the agent going when no X
       * connection is available.
       */

      if (nxagentDisplay &&
              nxagentDisplayBackup &&
                  (nxagentDisplay != nxagentDisplayBackup))
      {
        XCloseDisplay(nxagentDisplay);
      }
    case NOTHING:
      nxagentDisplay = nxagentDisplayBackup;
      break;
    default:
      FatalError("nxagentDisconnectDisplay: Display is in unknown state. Can't continue.");
  }

  reconnectDisplayState = NOTHING;
}

static Bool reloadVisuals(int flexibility)
{
  int i, nitems;
  XVisualInfo *newVisuals, *vi;
  unsigned long mask = VisualAllMask & ~VisualIDMask;

  newVisuals = malloc(sizeof(XVisualInfo) * nxagentNumVisuals);

  for (i = 0; i < nxagentNumVisuals; i++)
  {
    if ((vi = XGetVisualInfo(nxagentDisplay, mask, nxagentVisuals + i, &nitems)) == NULL)
    {
      #ifdef WARNING
      fprintf(stderr, "reloadVisual: failed to match this visual:\n");
      fprintf(stderr, "\tdepth = %d\n", nxagentVisuals[i].depth);
      fprintf(stderr, "\tclass = %d\n", nxagentVisuals[i].class);
      fprintf(stderr, "\tmask = (%ld,%ld,%ld)\n",
                nxagentVisuals[i].red_mask,
                  nxagentVisuals[i].green_mask,
                    nxagentVisuals[i].blue_mask);
      fprintf(stderr, "\tcolormap size = %d\n", nxagentVisuals[i].colormap_size);
      fprintf(stderr, "\tbits_per_rgb = %d\n", nxagentVisuals[i].bits_per_rgb);
      #endif

      if (nxagentVisuals[i].class == DirectColor)
      {
        #ifdef WARNING
        fprintf(stderr, "Warning: ignoring not matched DirectColor visual.\n");
        #endif

        nxagentVisualHasBeenIgnored[i] = TRUE;
        memcpy(newVisuals + i, nxagentVisuals + i, sizeof(XVisualInfo));
      }
      else
      {
        break;
      }
    }
    else
    {
      memcpy(newVisuals + i, vi, sizeof(XVisualInfo));
      nxagentVisualHasBeenIgnored[i] = FALSE;
      XFree(vi);
    }
  }

  if (i == nxagentNumVisuals)
  {
    nxagentVisuals = newVisuals;

    return TRUE;
  }

  free(newVisuals);

  return FALSE;
}

static Bool checkForColormapCompatibility(int flexibility)
{
  return (nxagentNumDefaultColormaps == nxagentNumDefaultColormapsRecBackup);
}

static Bool checkForDepthCompatibility(int flexibility)
{
  int i;

  if (nxagentNumDepths != nxagentNumDepthsRecBackup)
    return False;

  for (i = 0; i < nxagentNumDepths; i++)
    if (nxagentDepths[i] != nxagentDepthsRecBackup[i])
      return False;

  return True;
}

static Bool checkForPixmapFormatCompatibility(int flexibility)
{
  int i;

  if (nxagentNumPixmapFormats != nxagentNumPixmapFormatsRecBackup)
    return False;

  for (i = 0; i < nxagentNumPixmapFormats; i++)
    if ( (nxagentPixmapFormats[i].depth != nxagentPixmapFormatsRecBackup[i].depth) ||
         (nxagentPixmapFormats[i].bits_per_pixel != nxagentPixmapFormatsRecBackup[i].bits_per_pixel) ||
         (nxagentPixmapFormats[i].scanline_pad != nxagentPixmapFormatsRecBackup[i].scanline_pad) )
      return False;

  return True;
}

Bool nxagentReconnectDisplay(void *p0)
{
  XVisualInfo vi;
  int i;
  int flexibility = *(int*)p0;
  int packMethod, packQuality;

  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_DISPLAY_DEBUG)
  fprintf(stderr, "nxagentReconnectDisplay\n");
  #endif

  if (reconnectDisplayState)
  {
    fprintf(stderr, "nxagentReconnectDisplay: "
                    "trying to reconnect a session uncleanly disconnected\n");
    return False;
  }

  XSetIOErrorHandler(nxagentIOErrorHandler);

  /* We change the state of these globals parameter to initial value. */
  packMethod = nxagentPackMethod;
  packQuality = nxagentPackQuality;
  nxagentPackMethod = -1;
  nxagentPackQuality = -1;
  nxagentTightThreshold = -1;
  nxagentStopImgSz = -1;
  nxagentStopKarmaSz = -1;

  /*
   * FIXME: Should read value of the new display here.
   */

  nxagentDisplay = nxagentSafeOpenDisplay(nxagentDisplayName);

  if (nxagentDisplay == NULL)
  {
    nxagentSetReconnectError(FAILED_RESUME_DISPLAY_ALERT,
                                 "Couldn't open the display.");

    return FALSE;
  }

  nxagentAddXConnection();

  /*
   * Display is now open.
   */

  reconnectDisplayState = OPENED;

#ifdef NXAGENT_TIMESTAMP
  fprintf(stderr, "Display: display on real X opened, time is [%d] milliseconds.\n",
          GetTimeInMillis() - startTime);
#endif

  nxagentUseNXTrans = nxagentPostProcessArgs(nxagentDisplayName, nxagentDisplay,
                                                 DefaultScreenOfDisplay(nxagentDisplay));

  /*
   * FIXME: In theory we should free the image cache only
   * when the pack method or the set of visuals is changed,
   * but unfortunately it seems that there are still some
   * problems.
   */

  NXFreeCache();

  if (nxagentSynchronize)
  {
    XSynchronize(nxagentDisplay, True);
  }

  if (reloadVisuals(flexibility) == FALSE)
  {
    nxagentSetReconnectError(FAILED_RESUME_VISUALS_ALERT, 
                                 "Couldn't restore the required visuals.");

    return FALSE;
  }

  /* Display state is now GOT_VISUAL_INFO. */
  reconnectDisplayState = GOT_VISUAL_INFO;

  if (nxagentUserDefaultClass || nxagentUserDefaultDepth)
  {
    nxagentDefaultVisualIndex = UNDEFINED;
    for (i = 0; i < nxagentNumVisuals; i++)
      if ((!nxagentUserDefaultClass ||
           nxagentVisuals[i].class == nxagentDefaultClass) &&
          (!nxagentUserDefaultDepth ||
           nxagentVisuals[i].depth == nxagentDefaultDepth))
      {
        nxagentDefaultVisualIndex = i;
        break;
      }
    if (nxagentDefaultVisualIndex == UNDEFINED)
      FatalError("Unable to find desired default visual.\n");
  }
  else
  {
    vi.visualid = XVisualIDFromVisual(DefaultVisual(nxagentDisplay, DefaultScreen(nxagentDisplay)));
    nxagentDefaultVisualIndex = 0;
    for (i = 0; i < nxagentNumVisuals; i++)
      if (vi.visualid == nxagentVisuals[i].visualid)
        nxagentDefaultVisualIndex = i;
  }

  nxagentNumDefaultColormaps = nxagentNumVisuals;
  nxagentDefaultColormaps = (Colormap *) xrealloc(nxagentDefaultColormaps, nxagentNumDefaultColormaps * sizeof(Colormap));

  if (nxagentDefaultColormaps == NULL)
  {
    FatalError("xrealloc failed\n");
  }

  /* Display state is now ALLOC_DEF_COLORMAP. */
  reconnectDisplayState = ALLOC_DEF_COLORMAP;
  for (i = 0; i < nxagentNumDefaultColormaps; i++)
  {
    if (nxagentVisualHasBeenIgnored[i])
    {
      nxagentDefaultColormaps[i] = (XID)0;
    }
    else
    {
      nxagentDefaultColormaps[i] = XCreateColormap(nxagentDisplay,
                                                   DefaultRootWindow(nxagentDisplay),
                                                   nxagentVisuals[i].visual,
                                                   AllocNone);
    }
  }
  if (!checkForColormapCompatibility(flexibility))
  {
    nxagentSetReconnectError(FAILED_RESUME_COLORMAPS_ALERT,
                                 "Number of colormaps has changed.");

    return False;
  }

  nxagentDepths = XListDepths(nxagentDisplay, DefaultScreen(nxagentDisplay),
                            &nxagentNumDepths);

  /* Display state is now GOT_DEPTH_LIST. */
  reconnectDisplayState = GOT_DEPTH_LIST;

  if (!checkForDepthCompatibility(flexibility))
  {
    nxagentSetReconnectError(FAILED_RESUME_DEPTHS_ALERT, 
                                 "Couldn't restore all the required depths.");

    return False;
  }

  nxagentPixmapFormats = XListPixmapFormats(nxagentDisplay,
                                          &nxagentNumPixmapFormats);

  /* Display state is now GOT_PIXMAP_FORMAT_LIST. */
  reconnectDisplayState = GOT_PIXMAP_FORMAT_LIST;

  if (!checkForPixmapFormatCompatibility(flexibility))
  {
    nxagentSetReconnectError(FAILED_RESUME_PIXMAPS_ALERT,
                                 "Couldn't restore all the required pixmap formats.");

    return False;
  }

  #ifdef RENDER
  if (nxagentRenderEnable)
  {
    nxagentRenderExtensionInit();
  }

  if (nxagentRenderEnableRecBackup != nxagentRenderEnable)
  {
    nxagentRenderEnable = nxagentRenderEnableRecBackup;

    nxagentSetReconnectError(FAILED_RESUME_RENDER_ALERT,
                                 "Render extension not available or incompatible version.");

    return False;
  }
  #endif


  nxagentBlackPixel = BlackPixel(nxagentDisplay, DefaultScreen(nxagentDisplay));
  nxagentWhitePixel = WhitePixel(nxagentDisplay, DefaultScreen(nxagentDisplay));

  /*
   * Initialize the agent's event mask that will be requested
   * for the root or all the top level windows. If the nested
   * window is a child of an existing window we will need to
   * receive StructureNotify events. If we are going to manage
   * the changes in root window's visibility we'll also need
   * VisibilityChange events.
   */

  nxagentInitDefaultEventMask();

  /*
   * Create a pixmap for each depth matching the
   * local supported formats with format available
   * on the remote display.
   */

  nxagentSetDefaultDrawables();

  nxagentBitmapGC = XCreateGC(nxagentDisplay, nxagentDefaultDrawables[1], 0L, NULL);

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectDisplay: Going to create agent's confine window.\n");
  #endif

  nxagentConfineWindow = XCreateWindow(nxagentDisplay,
                                     DefaultRootWindow(nxagentDisplay),
                                     0, 0, 1, 1, 0, 0,
                                     InputOnly,
                                     CopyFromParent,
                                     0L, NULL);

  #ifdef TEST
  fprintf(stderr, "nxagentReconnectDisplay: Created agent's confine window with id [%ld].\n",
              nxagentConfineWindow);
  #endif

  nxagentLogoDepth = DefaultDepth(nxagentDisplay, DefaultScreen(nxagentDisplay));

  pV = nxagentVisuals[nxagentDefaultVisualIndex];

  r = pV.red_mask;
  g = pV.green_mask;
  b = pV.blue_mask;
  if(!pV.red_mask|| !pV.green_mask|| !pV.blue_mask)
  {
    nxagentLogoBlack = 0x000000;
    nxagentLogoRed   = 0xff0000;
    nxagentLogoWhite = 0xffffff;
  }
  else
  {
    for(or=0,off=0x800000;(r&(off>>or)) == 0; or++);
    for(og=0,off=0x800000;(g&(off>>og)) == 0; og++);
    for(ob=0,off=0x800000;(b&(off>>ob)) == 0; ob++);
    nxagentLogoRed   = nxagentLogoColor(0xff0000);
    nxagentLogoBlack = nxagentLogoColor(0x000000);
    nxagentLogoWhite = 0xffffff;
  }

#ifdef NXAGENT_LOGO_DEBUG
  fprintf(stderr, "mask r: %x g: %x b: %x\n", r, g, b);
  fprintf(stderr, "offset r: %d g: %d b: %d\n", or, og, ob);
#endif

  useXpmIcon = nxagentMakeIcon(nxagentDisplay, &nxagentIconPixmap, &nxagentIconShape);

#ifndef NXAGENT_NEVERSS
  nxagentScreenSaverPixmap =
    XCreatePixmapFromBitmapData(nxagentDisplay,
                                DefaultRootWindow(nxagentDisplay),
                                (char *)screensaver_bits,
                                screensaver_width,
                                screensaver_height,
                                nxagentWhitePixel,
                                nxagentBlackPixel,
                                DefaultDepth(nxagentDisplay,
                                DefaultScreen(nxagentDisplay)));
#endif
  /* Display state is now EVERYTHING_DONE. */
  reconnectDisplayState = EVERYTHING_DONE;

  return TRUE;
}

void nxagentAddXConnection()
{
  int fd = XConnectionNumber(nxagentDisplay);

  #ifdef WARNING
  fprintf(stderr, "Info: Adding the X connection %d to the device set.\n",
              fd);
  #endif

  AddEnabledDevice(fd);

  nxagentXConnectionNumber = fd;
}

void nxagentBreakDisplay()
{
  #ifdef WARNING
  fprintf(stderr, "Info: Going to shutdown the X connection %d.\n",
            nxagentXConnectionNumber);
  #endif

  /*
   * Using a shutdown() may not work, for example
   * if NX is using the memory to memory transport.
   */

  if (NXTransRunning())
  {
    NXTransDestroy();
  }
  else
  {
    shutdown(nxagentXConnectionNumber, SHUT_RDWR);
  }

  XSync(nxagentDisplay, False);
}

void nxagentRemoveXConnection()
{
  #ifdef WARNING
  fprintf(stderr, "Warning: Removing the X connection %d.\n",
            nxagentXConnectionNumber);
  #endif

  RemoveEnabledDevice(nxagentXConnectionNumber);
}

