/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2005 NoMachine, http://www.nomachine.com.          */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

/*

Copyright 1993 by Davor Matic

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation.  Davor Matic makes no representations about
the suitability of this software for any purpose.  It is provided "as
is" without express or implied warranty.

*/

#include "scrnintstr.h"
#include "font.h"
#include "fontstruct.h"
#include "misc.h"
#include "miscstruct.h"
#include "opaque.h"

#include "Agent.h"

#include "Display.h"
#include "Font.h"
#include "Error.h"

#include <stdio.h>
#include <sys/stat.h>
#include "resource.h"
#include "Reconnect.h"

#include "NXalert.h"
#include "Args.h"

#include NXAGENT_NXLIB_INCLUDE

#include <string.h>
#include <stdlib.h>

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG

#define NXAGENT_DEFAULT_FONT_DIR    "/usr/X11R6/lib/X11/fonts"
#define NXAGENT_ALTERNATE_FONT_DIR  "/usr/share/X11/fonts"

#define NXAGENT_DEFAULT_FONT_PATH  \
"/usr/X11R6/lib/X11/fonts/misc/,/usr/X11R6/lib/X11/fonts/Speedo/,\
/usr/X11R6/lib/X11/fonts/Type1/,/usr/X11R6/lib/X11/fonts/75dpi/,\
/usr/X11R6/lib/X11/fonts/100dpi/,/usr/X11R6/lib/X11/fonts/TTF/"

#define NXAGENT_ALTERNATE_FONT_PATH  \
"/usr/share/X11/fonts/misc/,/usr/share/X11/fonts/Speedo/,\
/usr/share/X11/fonts/Type1/,/usr/share/X11/fonts/75dpi/,\
/usr/share/X11/fonts/100dpi/,/usr/share/X11/fonts/TTF/"

#undef NXAGENT_FONTCACHE_DEBUG
#undef NXAGENT_RECONNECT_FONT_DEBUG
#undef NXAGENT_FONTMATCH_DEBUG

static int reconnectFlexibility;

static void nxagentCleanCacheAfterReconnect(void);
static void nxagentFontReconnect(FontPtr, XID, pointer);
static XFontStruct *nxagentLoadBestQueryFont(Display* dpy, char *fontName);

RESTYPE RT_NX_FONT;

#ifdef NXAGENT_RECONNECT_FONT_DEBUG
static void printFontCacheDump(char*);
#endif

typedef struct _nxagentFontRec
{
  char *name;
  int  status;
} nxagentFontRec, *nxagentFontRecPtr;

typedef struct _nxagentFontList
{
  nxagentFontRecPtr *list;
  int lenght;
  int listSize;
} nxagentFontList, *nxagentFontListPtr;

#ifdef NXAGENT_FONTEXCLUDE
char *excludeList[] = {"-ult1mo",(char*)NULL};
#endif

nxagentFontList nxagentRemoteFontList = {NULL, (int)0, (int)0};

int nxagentFontPrivateIndex;

typedef struct _nxCacheFontEntry
{
  Atom atom;
  XFontStruct *font_struct;
  char *name;
} nxCacheFontEntryRec, *nxCacheFontEntryRecPtr;

static struct _nxagentFontCache
{
  nxCacheFontEntryRecPtr *entry;
  int index;
  int size;
} nxagentFontCache = { NULL, (int) 0, (int) 0 };

#define CACHE_ENTRY_PTR (nxagentFontCache.entry)
#define CACHE_INDEX (nxagentFontCache.index)
#define CACHE_SIZE (nxagentFontCache.size)
#define CACHE_ENTRY(A) (CACHE_ENTRY_PTR[A])
#define CACHE_FSTRUCT(A) (CACHE_ENTRY(A) -> font_struct)
#define CACHE_NAME(A) (CACHE_ENTRY(A) -> name)
#define CACHE_ATOM(A) (CACHE_ENTRY(A) -> atom)

#define NXAGENT_FONT_SERVER_PATH "tcp/localhost:"

/*
 * This is used if nxagentFullGeneration is true
 * in CloseDisplay().
 */

void nxagentFreeFontCache(void)
{
  int i;

  #ifdef NXAGENT_FONTCACHE_DEBUG
  fprintf(stderr, "Font: Freeing nxagent font cache\n");
  #endif

  if (CACHE_INDEX == 0)
    return;

  #ifdef NXAGENT_FONTCACHE_DEBUG
  fprintf(stderr, "Font: Freeing nxagent font cache, there are [%d] entries.\n", CACHE_INDEX);
  #endif

  for (i = 0; i < CACHE_INDEX; i++)
  {
    #ifdef NXAGENT_FONTCACHE_DEBUG
    fprintf(stderr, "Font: Freeing nxagent font cache entry [%d] entry pointer is [%p], name [%s]\n",
                i, CACHE_ENTRY(i), CACHE_NAME(i));
    #endif

    if (CACHE_FSTRUCT(i))
    {
      XFreeFont(nxagentDisplay, CACHE_FSTRUCT(i));
    }

    xfree(CACHE_NAME(i));
    xfree(CACHE_ENTRY(i));
  }

  xfree(CACHE_ENTRY_PTR);
  CACHE_ENTRY_PTR = NULL;
  CACHE_INDEX = 0;
  CACHE_SIZE = 0;

  #ifdef NXAGENT_FONTCACHE_DEBUG
  fprintf(stderr, "Font: nxagent font cache fully freed\n");
  #endif

  return;
}

void nxagentListRemoteFonts(const char *searchPattern, const int maxNames)
{
  int i,j,q,pos;
  char **xList;
  int  xLen = 0;

  if (NXDisplayError(nxagentDisplay)) return;

  if (nxagentFontFind(searchPattern, &pos))
  {
     if (nxagentRemoteFontList.list[pos]->status >= maxNames)
     {
       #ifdef NXAGENT_FONTMATCH_DEBUG
       fprintf(stderr, "Font: nxagent remote list [%s] request [%d] elements status is [%d] elements.\n",
                   searchPattern,maxNames,nxagentRemoteFontList.list[pos]->status );
       #endif

       return;
     }
  }

  xList = XListFonts(nxagentDisplay,searchPattern, maxNames, &xLen);

  #ifdef NXAGENT_FONTMATCH_DEBUG
  fprintf(stderr, "Font: nxagent remote list [%s] has %d elements.\n", searchPattern, xLen);
  #endif

/* add the ListFont request pattern to the list with the last maxnames quantity requested */

  nxagentListRemoteAddName(searchPattern, maxNames);

  for (i = 0; i < xLen; i++)
  {
    q = 1;
    #ifdef NXAGENT_FONTEXCLUDE
    j = 0;
    while (excludeList[j])
    {
     if (strncasecmp(excludeList[j], xList[i],strlen(excludeList[j])) == 0)
     {
       q = 0;

       #ifdef NXAGENT_FONTMATCH_DEBUG
       fprintf(stderr, "Font: nxagent remote list [%s] excluded.\n", xList[i]);
       #endif

       break;
     }
     j++;
    }
    #endif /* NXAGENT_FONTEXCLUDE */
    nxagentListRemoteAddName(xList[i], q);
  }

  XFreeFontNames(xList);

  #ifdef NXAGENT_FONTMATCH_DEBUG_no
  fprintf(stderr, "Font: Printing remote font list.\n");
  for (i = 0; i < nxagentRemoteFontList.lenght; i++)
  {
    fprintf(stderr, "Font# %d, \"%s\"\n", i, nxagentRemoteFontList.list[i]->name);
  }
  fprintf(stderr, "Font: End of list\n");
  #endif
}

void nxagentListRemoteAddName(const char *name, int status)
{
int pos;
    if (nxagentFontFind(name, &pos))
    {
       if (nxagentRemoteFontList.list[pos]->status < status)
       {
         nxagentRemoteFontList.list[pos]->status = status;

         #ifdef NXAGENT_FONTMATCH_DEBUG
         fprintf(stderr, "Font: Font# %d, [%s] change status to %s\n",
                     pos, nxagentRemoteFontList.list[pos]->name,nxagentRemoteFontList.list[pos]->status?"OK":"deleted");
         #endif
       }
       return;
    }

    if (nxagentRemoteFontList.lenght == nxagentRemoteFontList.listSize)
    {
       nxagentRemoteFontList.list=xrealloc(nxagentRemoteFontList.list,sizeof(nxagentFontRecPtr) * (nxagentRemoteFontList.listSize + 1000));

       if (nxagentRemoteFontList.list == NULL)
       {
           FatalError("Font: remote list memory re-allocation failed!.\n");
       }

       nxagentRemoteFontList.listSize += 1000;
    }

    if (pos < nxagentRemoteFontList.lenght)
    {
      #ifdef NXAGENT_FONTMATCH_DEBUG
      fprintf(stderr, "Font: Going to move list from %p to %p len = %d!.\n",
                  &nxagentRemoteFontList.list[pos], &nxagentRemoteFontList.list[pos+1],
                      (nxagentRemoteFontList.lenght - pos) * sizeof(nxagentFontRecPtr));
      #endif

      memmove(&nxagentRemoteFontList.list[pos+1],
                  &nxagentRemoteFontList.list[pos],
                      (nxagentRemoteFontList.lenght - pos) * sizeof(nxagentFontRecPtr));
    }

    if ((nxagentRemoteFontList.list[pos] = xalloc(sizeof(nxagentFontRec))))
    {
      nxagentRemoteFontList.list[pos]->name = xalloc(strlen(name) +1);
      if (nxagentRemoteFontList.list[pos]->name == NULL)
      {
         fprintf(stderr, "Font: remote list name memory allocation failed!.\n");
         return;
      }
    }
    else
    {
       fprintf(stderr, "Font: remote list record memory allocation failed!.\n");
       return;
    }
    strcpy(nxagentRemoteFontList.list[pos]->name,name);
    nxagentRemoteFontList.list[pos]->status = status;
    nxagentRemoteFontList.lenght++;

    #ifdef NXAGENT_FONTMATCH_DEBUG
    fprintf(stderr, "Font: remote font list added [%s] in position [%d] as %s !.\n",
                name, pos, status ? "OK" : "deleted");
    fprintf(stderr, "Font: remote font list total len is [%d] Size is [%d] !.\n",
                nxagentRemoteFontList.lenght, nxagentRemoteFontList.listSize);
    #endif
}

Bool nxagentFontFind(const char *name, int *pos)
{
 int low,high,res,iter,lpos;

 if (!nxagentRemoteFontList.lenght)
 {
    *pos=0;
    return False;
 }
 low = 0;
 high = nxagentRemoteFontList.lenght - 1;
 iter = 0;
 res = 1;
 lpos = nxagentRemoteFontList.lenght;
 while (low <= high )
 {
   *pos = (high + low)/2;
   iter ++;
   res = strcasecmp(nxagentRemoteFontList.list[*pos]->name,name);
   if (res > 0)
   {
      high = *pos - 1;
      lpos = *pos;
      continue;
   }
   else if (res < 0)
   {
      low = *pos + 1;
      lpos = low;
      continue;
   }
   break;
 }
 *pos = (res == 0)?*pos:lpos;

 #ifdef NXAGENT_FONTMATCH_DEBUG
 if (res == 0)
   fprintf(stderr, "Font: font found in %d iterations in pos = %d\n", iter, *pos);
 else
   fprintf(stderr, "Font: not font found in %d iterations insertion pos is = %d\n", iter, *pos);
 #endif

 return (res == 0);

}

Bool nxagentFontLookUp(const char *name)
{
  int i;
  if (name)
    if (!strlen(name))
       return 0;
  if (nxagentFontFind(name, &i))
    return (nxagentRemoteFontList.list[i]->status > 0);
  else
    return 0;
}

Bool nxagentRealizeFont(ScreenPtr pScreen, FontPtr pFont)
{
  pointer priv;
  Atom name_atom, value_atom;
  int nprops;
  FontPropPtr props;
  int i;
  char *name;
  char *origName = (char*) pScreen;

  FontSetPrivate(pFont, nxagentFontPrivateIndex, NULL);

  if (requestingClient && XpClientIsPrintClient(requestingClient, NULL))
    return True;

  name_atom = MakeAtom("FONT", 4, True);
  value_atom = 0L;

  nprops = pFont->info.nprops;
  props = pFont->info.props;

  for (i = 0; i < nprops; i++)
    if ((Atom)props[i].name == name_atom) {
      value_atom = props[i].value;
      break;
    }

  if (!value_atom) return False;

  name = (char *)NameForAtom(value_atom);

  #ifdef NXAGENT_FONTCACHE_DEBUG
  fprintf(stderr, "Font: nxagentRealizeFont, realizing font: %s\n", validateString(name));
  fprintf(stderr, "                                 atom: %ld\n", value_atom);
  fprintf(stderr, "Font: Cache dump:\n");
  for (i = 0; i < CACHE_INDEX; i++)
  {
      fprintf(stderr, "nxagentFontCache.entry[%d]->name: %s font_struct at %p\n",
                  i, CACHE_NAME(i), CACHE_FSTRUCT(i));
  }
  #endif

  if (!name) return False;

  if ((strcasecmp(origName, name) != 0) && !strchr(origName,'*'))
  {
     #ifdef NXAGENT_FONTMATCH_DEBUG
     fprintf(stderr, "Font: Changing font name to realize from [%s] to [%s]\n",
                 validateString(name), origName);
     #endif

     name = origName;
  }

  priv = (pointer)xalloc(sizeof(nxagentPrivFont));
  FontSetPrivate(pFont, nxagentFontPrivateIndex, priv);

  nxagentFontPriv(pFont) -> mirrorID = 0;

  for (i = 0; i < nxagentFontCache.index; i++)
  {
/*      if (value_atom == CACHE_ATOM(i))*/
     if (strcasecmp(CACHE_NAME(i), name) == 0)
     {
        #ifdef NXAGENT_FONTCACHE_DEBUG
        fprintf(stderr, "Font: nxagentFontCache hit [%s] = [%s]!\n", CACHE_NAME(i), validateString(name));
        #endif

        break;
     }
  }

  if (i < CACHE_INDEX)
  {
      nxagentFontPriv(pFont)->font_struct = CACHE_FSTRUCT(i);
      strcpy(nxagentFontPriv(pFont)->fontName, name);
  }
  else if (!NXDisplayError(nxagentDisplay))
  {
      #ifdef NXAGENT_FONTCACHE_DEBUG
      fprintf(stderr, "Font: nxagentFontCache fail.\n");
      #endif

      if (CACHE_INDEX == CACHE_SIZE)
      {
        CACHE_ENTRY_PTR = xrealloc(CACHE_ENTRY_PTR, sizeof(nxCacheFontEntryRecPtr) * (CACHE_SIZE + 100));

        if (CACHE_ENTRY_PTR == NULL)
        {
           FatalError("Font: cache list memory re-allocation failed!.\n");
        }

        CACHE_SIZE += 100;
     }

     CACHE_ENTRY(CACHE_INDEX) = xalloc(sizeof(nxCacheFontEntryRec));

     if (CACHE_ENTRY(CACHE_INDEX) == NULL)
     {
        return False;
/*        nxagentExitHandler("Font: Memory allocation error in nxagent font cache code (#1)");*/
     }

     CACHE_NAME(CACHE_INDEX) = xalloc(strlen(name) + 1);

     if (CACHE_NAME(CACHE_INDEX) == NULL)
     {
        return False;
     }

     #ifdef NXAGENT_FONTMATCH_DEBUG
     fprintf(stderr, "Font: going to realize font [%s],[%s] on real X server.\n", validateString(name), origName);
     #endif

     nxagentFontPriv(pFont)->font_struct = XLoadQueryFont(nxagentDisplay, name);
     strcpy(nxagentFontPriv(pFont)->fontName, name);
     if (nxagentFontPriv(pFont)->font_struct != NULL)
     {
       CACHE_ATOM(i) = value_atom;
       strcpy(CACHE_NAME(i), name);
       CACHE_FSTRUCT(i) = nxagentFontPriv(pFont)->font_struct;
       CACHE_INDEX++;

       nxagentFontPriv(pFont) -> mirrorID = FakeClientID(0);
       AddResource(nxagentFontPriv(pFont) -> mirrorID, RT_NX_FONT, pFont);

       #ifdef NXAGENT_FONTCACHE_DEBUG
       fprintf(stderr, "Font: nxagentFontCache adds font [%s] in pos. [%d].\n",
                   validateString(name), CACHE_INDEX - 1);
       #endif
     }
  }
  else
  {
    nxagentFontPriv(pFont)->font_struct = NULL;
    strcpy(nxagentFontPriv(pFont)->fontName, name);

    #ifdef WARNING
    fprintf(stderr, "Font: failed to load font %s from cache when X connection is not available\n",
                validateString(name));
    #endif

    return False;
  }

  #ifdef NXAGENT_FONTMATCH_DEBUG
  if (nxagentFontPriv(pFont)->font_struct == NULL)
  {
    if (nxagentFontLookUp(name) == False)
    {
      fprintf(stderr, "Font: nxagentRealizeFont failed with font Font=%s, not in our remote list\n",
                  validateString(name));
    }
    else
    {
      fprintf(stderr, "Font: nxagentRealizeFont failed with font Font=%s but the font is in our remote list\n",
                  validateString(name));
    }
  }
  else
      fprintf(stderr, "Font: nxagentRealizeFont OK realizing font Font=%s\n",
                  validateString(name));
  #endif

  return (nxagentFontPriv(pFont)->font_struct != NULL);
}

Bool nxagentUnrealizeFont(ScreenPtr pScreen, FontPtr pFont)
{
  if (nxagentFontPriv(pFont))
  {
    if (!NXDisplayError(nxagentDisplay))
    {
      if (nxagentFontStruct(pFont))
      {
         int i;

         for (i = 0; i < CACHE_INDEX; i++)
         {
           /*if (CACHE_FSTRUCT(i)->fid == nxagentFont(pFont))*/
           if(CACHE_FSTRUCT(i) == nxagentFontStruct(pFont))
           {
             #ifdef NXAGENT_FONTCACHE_DEBUG
             fprintf(stderr, "Font: nxagentUnrealizeFont not freeing font in cache.\n");
             #endif

             break;
           }
         }

         if (i == CACHE_INDEX)
         {
           /* this font is not cached. */
           #ifdef NXAGENT_FONTCACHE_DEBUG
           fprintf(stderr, "Font: nxagentUnrealizeFont freeing font not in cache: %d\n",
                       CACHE_ATOM(i));
           #endif

           XFreeFont(nxagentDisplay, nxagentFontStruct(pFont));
         }
       }
    }

    if (nxagentFontPriv(pFont) -> mirrorID)
      FreeResource(nxagentFontPriv(pFont) -> mirrorID, RT_NONE);

    xfree(nxagentFontPriv(pFont));
    FontSetPrivate(pFont, nxagentFontPrivateIndex, NULL);
  }

  return True;
}

int nxagentDestroyNewFontResourceType(pointer p, XID id)
{
  #ifdef TEST
  fprintf(stderr, "nxagentDestroyNewFontResourceType: Destroying mirror id [%ld] for font at [%p].\n",
              nxagentFontPriv((FontPtr) p) -> mirrorID, (void *) p);
  #endif

  nxagentFontPriv((FontPtr) p) -> mirrorID = None;

  return 1;
}

static XFontStruct *nxagentLoadBestQueryFont(Display* dpy, char *fontName)
{
  XFontStruct *fontStruct;

  #ifdef WARNING
  fprintf(stderr, "nxagentFontReconnect: WARNING! Failed to load font '%s'. Replacing with '%s'.\n",
              fontName, "fixed");
  #endif

  fontStruct = XLoadQueryFont(dpy, "fixed");

  return fontStruct;
}

static void nxagentFontDisconnect(FontPtr pFont, XID param1, pointer param2)
{
  nxagentPrivFont *privFont;
  Bool *pBool = (Bool*)param2;
  int i;

  if (pFont == NULL || !*pBool)
    return;

  privFont = nxagentFontPriv(pFont);

  #ifdef NXAGENT_RECONNECT_FONT_DEBUG
  fprintf(stderr, "nxagentFontDisconnect: pFont %x, XID %lx\n",
              pFont, privFont -> font_struct ? nxagentFont(pFont) : 0);
  #endif

  for (i = 0; i < CACHE_INDEX; i++)
  {
    if (strcasecmp(CACHE_NAME(i), privFont -> fontName) == 0)
    {
      #ifdef NXAGENT_RECONNECT_FONT_DEBUG
      fprintf(stderr, "nxagentFontDisconnect: font %s found in cache at position %d\n",
              privFont -> fontName, i);
      #endif

      privFont -> font_struct = NULL;
      return;
    }
  }

  #ifdef NXAGENT_RECONNECT_FONT_DEBUG
  fprintf(stderr, "nxagentFontDisconnect: WARNING font %s not found in cache freeing it now\n",
              privFont -> fontName);
  #endif

  if (privFont -> font_struct)
  {
     XFreeFont(nxagentDisplay, privFont -> font_struct);
     privFont -> font_struct = NULL;
  }
}

static void nxagentFontReconnect(FontPtr pFont, XID param1, pointer param2)
{
  int i;
  nxagentPrivFont *privFont;
  Bool *pBool = (Bool*)param2;

  if (pFont == NULL || !*pBool)
    return;

  privFont = nxagentFontPriv(pFont);

  #ifdef NXAGENT_RECONNECT_FONT_DEBUG
  fprintf(stderr, "nxagentFontReconnect: pFont %p - XID %lx - name %s\n",
              (void*) pFont, (privFont -> font_struct) ? nxagentFont(pFont) : 0,
                  privFont -> fontName);
  #endif

  for (i = 0; i < CACHE_INDEX; i++)
  {
    if (strcasecmp(CACHE_NAME(i), privFont -> fontName) == 0)
    {
      #ifdef NXAGENT_RECONNECT_FONT_DEBUG
      fprintf(stderr, "\tfound in cache");
      #endif

      if (!CACHE_FSTRUCT(i))
      {
        #ifdef NXAGENT_RECONNECT_FONT_DEBUG
        fprintf(stderr, " --- font struct not valid\n");
        #endif

        break;
      }

      nxagentFontStruct(pFont) = CACHE_FSTRUCT(i);

      return;
    }
  }

  if (i == CACHE_INDEX)
  {
    FatalError("nxagentFontReconnect: font not found in cache.");
  }

  privFont -> font_struct = XLoadQueryFont(nxagentDisplay, privFont -> fontName);

  if ((privFont -> font_struct == NULL) && reconnectFlexibility)
  {
    privFont -> font_struct = nxagentLoadBestQueryFont(nxagentDisplay, privFont -> fontName);
  }

  if (privFont->font_struct != NULL)
  {
    #ifdef NXAGENT_RECONNECT_FONT_DEBUG
    fprintf(stderr, "\tXID %lx\n", privFont -> font_struct -> fid);
    #endif

    CACHE_FSTRUCT(i) = privFont -> font_struct;
  }
  else
  {
    #ifdef NXAGENT_RECONNECT_FONT_DEBUG
    fprintf(stderr, "nxagentFontReconnect: failed\n");
    #endif

    nxagentSetReconnectError(FAILED_RESUME_FONTS_ALERT,
                                 "Couldn't restore the font '%s'", privFont -> fontName);

    *pBool = False;
  }

  return;
}

static void nxagentFreeCacheBeforeReconnect(void)
{
  int i;

  #ifdef NXAGENT_RECONNECT_FONT_DEBUG
  printFontCacheDump("nxagentFreeCacheBeforeReconnect");
  #endif

  for (i = 0; i < CACHE_INDEX; i++)
  {
    if (CACHE_FSTRUCT(i))
    {
      XFreeFont(nxagentDisplay, CACHE_FSTRUCT(i));

      CACHE_FSTRUCT(i) = NULL;
    }
  }
}

static void nxagentCleanCacheAfterReconnect(void)
{
  int i, j;
  int real_size = CACHE_INDEX;
  nxCacheFontEntryRecPtr swapEntryPtr;

  #ifdef NXAGENT_RECONNECT_FONT_DEBUG
  printFontCacheDump("nxagentCleanCacheAfterReconnect");
  #endif

  for (i = 0; i < CACHE_INDEX; i++)
  {
    if( CACHE_FSTRUCT(i) == NULL)
    {
      XFree(CACHE_NAME(i));
      real_size--;
    }
  }

  for (i = 0; i < real_size; i++)
  {
      /* Find - first bad occurrence if exist. */
      while ((i < real_size) && CACHE_FSTRUCT(i)) i++;

      /* Really nothing more to do. */
      if (i == real_size)
        break;

      /*
       * Find - first good occurrence (moving backward from right end) entry in
       *        order to replace the bad one.
       */
      for (j = CACHE_INDEX - 1; CACHE_FSTRUCT(j) == NULL; j--);

      /*
       * Now we can swap the two entry
       * and reduce the Cache index
       */
      swapEntryPtr = CACHE_ENTRY(i);
      CACHE_ENTRY(i) = CACHE_ENTRY(j);
      CACHE_ENTRY(j) = swapEntryPtr;
  }

  CACHE_INDEX = real_size;
}

#ifdef NXAGENT_RECONNECT_FONT_DEBUG
static void printFontCacheDump(char* msg)
{
  int i;

  fprintf(stderr, "%s - begin -\n", msg);

  for (i = 0; i < CACHE_INDEX; i++)
  {
    if (CACHE_FSTRUCT(i))
    {
      fprintf(stderr, "\tXID %lx - %s\n", CACHE_FSTRUCT(i) -> fid, CACHE_NAME(i));
    }
    else
    {
      fprintf(stderr, "\tdestroyed   - %s\n", CACHE_NAME(i));
    }
  }
  fprintf(stderr, "%s - end   -\n", msg);
}
#endif

Bool nxagentReconnectAllFonts(void *p0)
{
  int cid;
  Bool fontSuccess = True;

  reconnectFlexibility = *((int *) p0);

  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_FONT_DEBUG)
  fprintf(stderr, "nxagentReconnectAllFonts\n");
  #endif

  for (cid = 0; cid < MAXCLIENTS; cid++)
  {
    if (clients[cid] && fontSuccess)
    {
      FindClientResourcesByType(clients[cid], RT_NX_FONT,
                                    (FindResType) nxagentFontReconnect, &fontSuccess);
    }
  }

  if (fontSuccess)
  {
    nxagentCleanCacheAfterReconnect();
  }

  return fontSuccess;
}

Bool nxagentDisconnectAllFonts()
{
  int cid;
  Bool fontSuccess = True;

  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_FONT_DEBUG)
  fprintf(stderr, "nxagentDisconnectAllFonts\n");
  #endif

  nxagentFreeCacheBeforeReconnect();

  for(cid = 0; cid < MAXCLIENTS; cid++)
  {
    if( clients[cid] && fontSuccess )
    {
      FindClientResourcesByType(clients[cid], RT_NX_FONT,
                                    (FindResType) nxagentFontDisconnect, &fontSuccess);
    }
  }

  return True;
}

static char *nxagentFontServerPath(char *displayName)
{
  char *portStr, *comPortStr;
  char *fontServerPath;
  int port, comPort;

  if ((portStr = strrchr(displayName, ':')) == NULL)
  {
    #ifdef WARNING	  
    fprintf(stderr, "nxagentFontPath: bad display name.");
    #endif

    return NULL;
  }
  
  ++portStr;

  port = atoi(portStr);
  comPort = 10000 + port;
  
  if ((comPortStr = malloc(6)) == NULL)
  {
    FatalError("nxagentFontPath: malloc failed.");
  }
  
  sprintf(comPortStr, "%d", comPort);

  if ((fontServerPath = malloc(strlen(NXAGENT_FONT_SERVER_PATH) + strlen(comPortStr) + 1 )) == NULL)
  {
    FatalError("nxagentFontPath: malloc failed.");
  }

  strcpy(fontServerPath, NXAGENT_FONT_SERVER_PATH);
  strcat(fontServerPath, comPortStr);

  free(comPortStr);

  return fontServerPath;
}

static Bool nxagentInstallFontPath(char *fontPath)
{
  char **fontPaths, **localFontPaths, **newFontPaths;
  int nPaths;
  
  /* Get font Path from Xserver */
  fontPaths = XGetFontPath(nxagentDisplay, &nPaths);

  if ((newFontPaths =  malloc((nPaths + 1) * sizeof(char *))) == NULL)
  {
    FatalError("nxagentInstallFontPath: malloc failed.");
  }

  memcpy(newFontPaths, fontPaths, nPaths * sizeof(char*));

  localFontPaths = newFontPaths;
  localFontPaths += nPaths;
  *localFontPaths = fontPath;
  
  /* Set Font Path */
  XSetFontPath(nxagentDisplay, newFontPaths, nPaths  + 1);

  XFreeFontPath(fontPaths);
  free(newFontPaths);

  return True;
}

static Bool nxagentUninstallFontPath(char *fontPath)
{
  int counter = 0;
  int nPaths, k;
  char **fontPaths, **newFontPaths;
  char **localFontPaths, **localNewFontPaths;

  /* Get font path from the server */
  fontPaths = XGetFontPath(nxagentDisplay, &nPaths);

  if ((newFontPaths = (char **) malloc ((nPaths) * sizeof(char *))) == NULL)
  {
    FatalError("nxagentUninstallFontPath: malloc failed.");
  }

  localFontPaths = fontPaths;
  localNewFontPaths = newFontPaths;

  for (k = 0; k < nPaths; k++)
  {
    if (strcmp(*localFontPaths, fontPath) != 0)
    {
      *localNewFontPaths = *localFontPaths;
      localNewFontPaths++;
      counter++;
    }
    
    localFontPaths++;
  }

  XSetFontPath(nxagentDisplay, newFontPaths, counter);

  XFreeFontPath(fontPaths);
  free(newFontPaths);

  return True;
}

Bool nxagentInstallFontServerPath(void)
{
  char *fontServerPath;

  fontServerPath = nxagentFontServerPath(nxagentDisplayName);

  nxagentInstallFontPath(fontServerPath);

  free(fontServerPath);

  return True;
}

Bool nxagentUninstallFontServerPath(void)
{
  char *fontServerPath;

  fontServerPath = nxagentFontServerPath(nxagentDisplayName);

  nxagentUninstallFontPath(fontServerPath);

  free(fontServerPath);

  return True;
}

void nxagentVerifyDefaultFontPath(void)
{
  struct stat dirStat;

  if (stat(NXAGENT_DEFAULT_FONT_DIR, &dirStat) == 0 &&
          S_ISDIR(dirStat.st_mode) != 0)
  {
    /*
     * Let's use the old "/usr/X11R6/lib/X11/fonts" style.
     */

    #ifdef TEST
    fprintf(stderr, "nxagentVerifyDefaultFontPath: Assuming fonts in directory [%s].\n",
                validateString(NXAGENT_DEFAULT_FONT_DIR));
    #endif

    defaultFontPath = NXAGENT_DEFAULT_FONT_PATH;
  }
  else if (stat(NXAGENT_ALTERNATE_FONT_DIR, &dirStat) == 0)
  {
    /*
     * Let's use the new "/usr/share/X11/fonts" path.
     */

    #ifdef TEST
    fprintf(stderr, "nxagentVerifyDefaultFontPath: Assuming fonts in directory [%s].\n",
                validateString(NXAGENT_ALTERNATE_FONT_DIR));
    #endif

    defaultFontPath = NXAGENT_ALTERNATE_FONT_PATH;
  }
  else
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentVerifyDefaultFontPath: WARNING! Can't find a valid font directory.\n");

    fprintf(stderr, "nxagentVerifyDefaultFontPath: WARNING! Using font path [%s].\n",
                validateString(defaultFontPath));
    #endif
  }

  return;
}


