/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2005 NoMachine, http://www.nomachine.com.          */
/*                                                                        */
/* NXAGENT, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

/*

Copyright 1993 by Davor Matic

Permission to use, copy, modify, distribute, and sell this software
and its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation.  Davor Matic makes no representations about
the suitability of this software for any purpose.  It is provided "as
is" without express or implied warranty.

*/

#include "X.h"
#include "Xproto.h"
#include "gcstruct.h"
#include "windowstr.h"
#include "pixmapstr.h"
#include "scrnintstr.h"
#include "fontstruct.h"
#include "mistruct.h"
#include "region.h"

#include "Agent.h"

#include "Display.h"
#include "GC.h"
#include "GCOps.h"
#include "Drawable.h"
#include "Pixmap.h"
#include "Font.h"
#include "Colormap.h"

#include "../../fb/fb.h"

/*
 * Set here the required log level.
 */

#define PANIC
#define WARNING
#undef  TEST
#undef  DEBUG

#define GC_SAFEVECTOR_BLOCKSIZE 32 

GCPtr *nxagentGCSafeVector;
int nxagentGCSafeVectorSize;

int nxagentGCPrivateIndex;

#ifdef NXAGENT_GCOPTIMIZE
int nDiscarded;
#endif

static void nxagentGCReconnect(void*, XID, void*);
static void nxagentDummyChangeClip(GCPtr, int, pointer, int);

struct nxagentGCRec
{
  GCPtr pGC;
  XlibGC gc;
  struct nxagentGCRec *next;
};

static struct
{
  struct nxagentGCRec *first;
  struct nxagentGCRec *last;
  int size;
} nxagentGCList = { NULL, NULL, 0 };

static struct nxagentGCRec *nxagentGetFirstGC(void);

static void nxagentRestoreGCList(void);

static GCFuncs nxagentFuncs =
{
  nxagentValidateGC,
  nxagentChangeGC,
  nxagentCopyGC,
  nxagentDestroyGC,
  nxagentChangeClip,
  nxagentDestroyClip,
  nxagentCopyClip,
};

static GCOps nxagentOps =
{
  nxagentFillSpans,
  nxagentSetSpans,
  nxagentPutImage,
  nxagentCopyArea,
  nxagentCopyPlane,
  nxagentPolyPoint,
  nxagentPolylines,
  nxagentPolySegment,
  nxagentPolyRectangle,
  nxagentPolyArc,
  nxagentFillPolygon,
  nxagentPolyFillRect,
  nxagentPolyFillArc,
  nxagentPolyText8,
  nxagentPolyText16,
  nxagentImageText8,
  nxagentImageText16,
  nxagentImageGlyphBlt,
  nxagentPolyGlyphBlt,
  nxagentPushPixels
};

static void nxagentAddGCToSafeVector(GCPtr pGC)
{
  GCPtr *result;
  int i;

  /*
   * The nxagentInitGCSafeVector() can fail.
   */

  if (nxagentGCSafeVectorSize == 0)
  {
    return;
  }

  /* 
   * To create a short list some GC are ignored. The stipple pixmaps
   * are always automatically reconnected. 
   */

  if (pGC -> stipple)
  {
    return;
  }

  for (i = 0; i < nxagentGCSafeVectorSize; i++)
  {
    if (nxagentGCSafeVector[i] == pGC)
    {
      return;
    }
  }

  i = 0;

  while (nxagentGCSafeVector[i] != NULL && i < nxagentGCSafeVectorSize)
  {
    i++;
  }

  if (i >= nxagentGCSafeVectorSize)
  {
    result = (GCPtr *) realloc(nxagentGCSafeVector, (nxagentGCSafeVectorSize+GC_SAFEVECTOR_BLOCKSIZE) * sizeof(GCPtr));

    if (result == NULL)
    {
      #ifdef WARNING
      fprintf(stderr, "nxagentAddGCToSafeVector: realloc failed!");
      #endif

      return;
    }

    nxagentGCSafeVector = result;
    memset(&nxagentGCSafeVector[nxagentGCSafeVectorSize], 0, GC_SAFEVECTOR_BLOCKSIZE);
    nxagentGCSafeVectorSize += GC_SAFEVECTOR_BLOCKSIZE;
  }

  nxagentGCSafeVector[i] = pGC;

  #ifdef TEST
  fprintf(stderr, "nxagentAddGCToSafeVector: GC %p added to SafeVector.\n", (void *)pGC);
  #endif
}

void nxagentInitGCSafeVector()
{
  if ((nxagentGCSafeVector = (GCPtr *) calloc(GC_SAFEVECTOR_BLOCKSIZE, sizeof(GCPtr))) == NULL)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentInitGCSafeVector: calloc failed!\n");
    #endif 

    return; 
  }

  nxagentGCSafeVectorSize = GC_SAFEVECTOR_BLOCKSIZE;
}

void nxagentDeleteGCSafeRec(GCPtr pGC)
{
  int i;

  for (i = 0; i < nxagentGCSafeVectorSize; i++)
  {
    if (nxagentGCSafeVector[i] == pGC)
      nxagentGCSafeVector[i] = NULL;
  }
}

static void nxagentGCSafeVectorReconnect()
{
  int i;
  Bool success = True;

  for (i = 0; i < nxagentGCSafeVectorSize && success; i++)
  {
    if (nxagentGCSafeVector[i] != NULL)
    {
      nxagentGCReconnect(nxagentGCSafeVector[i], (XID)0, &success);
    }
  }
}

Bool nxagentCreateGC(GCPtr pGC)
{
  pGC->clientClipType = CT_NONE;
  pGC->clientClip = NULL;

  pGC->funcs = &nxagentFuncs;
  pGC->ops = &nxagentOps;

  pGC->miTranslate = 1;

  if (pGC -> stipple && !nxagentPixmapIsVirtual(pGC -> stipple))
  {
    #ifdef DEBUG
    fprintf(stderr, "nxagentCreateGC: GC at [%p] got real stipple at [%p] switched to virtual.\n",
                (void*)pGC, (void*)pGC -> stipple);
    #endif

    pGC -> stipple = nxagentVirtualPixmap(pGC -> stipple);
  }

  nxagentGCPriv(pGC)->gc = XCreateGC(nxagentDisplay,
                                   nxagentDefaultDrawables[pGC->depth],
                                   0L, NULL);

  nxagentAddGCToSafeVector(pGC);

#ifdef NXAGENT_MVFB
{
   FbGCPrivPtr pPriv;

#ifdef NXAGENT_MVFB_DEBUG
      fprintf(stderr, "nxagentCreateGC: GC [%lx]\n", pGC);
#endif

   pPriv = (pGC)->devPrivates[fbGCPrivateIndex].ptr;

   fbGetRotatedPixmap(pGC) = 0;
   fbGetExpose(pGC) = 1;
   fbGetFreeCompClip(pGC) = 0;
   fbGetCompositeClip(pGC) = 0;

   pPriv->bpp = BitsPerPixel (pGC->depth);
}
#endif

  nxagentGCPriv(pGC)->nClipRects = 0;
#ifdef NXAGENT_GCOPTIMIZE
  memset(&(nxagentGCPriv(pGC) -> lastServerValues),0,sizeof(XGCValues));
#endif
  return True;
}

void nxagentValidateGC(GCPtr pGC, unsigned long changes, DrawablePtr pDrawable)
{
  PixmapPtr lastTile, lastStipple;

  DrawablePtr pVirtual = (pDrawable -> type == DRAWABLE_PIXMAP) ?
                          nxagentVirtualDrawable(pDrawable) :
                          pDrawable;

  #ifdef TEST
  fprintf(stderr, "nxagentValidateGC: Going to validate GC at [%p] for drawable at [%p] with changes [%lx].\n",
              (void *) pGC, (void *) pDrawable, changes);
  #endif

  pGC->lastWinOrg.x = pDrawable->x;
  pGC->lastWinOrg.y = pDrawable->y;

#ifdef NXAGENT_MVFB


  /*
   * Before validate GC tile in framebuffer 
   * we have to be sure that this tile has 
   * correct pointer to virtual data.
   */

  if (!pGC -> tileIsPixel && !nxagentPixmapIsVirtual(pGC -> tile.pixmap))
  {
    pGC -> tile.pixmap = nxagentVirtualPixmap(pGC -> tile.pixmap); 
  }

  lastTile = pGC -> tile.pixmap;

  lastStipple = pGC->stipple;
  
  if (lastStipple)
    pGC->stipple = nxagentVirtualPixmap(pGC->stipple);

  #ifdef TEST
  fprintf(stderr, "nxagentValidateGC: Drawable at [%p] has type [%s] virtual [%p] bits per pixel [%d].\n",
              (void *) pDrawable, (pDrawable -> type == DRAWABLE_PIXMAP) ? "PIXMAP" : "WINDOW",
                  (void *) pVirtual, pVirtual -> bitsPerPixel);
  #endif

  if (pVirtual -> bitsPerPixel == 0)
  {
    /*
     * Don't enter fbValidateGC() with 0 bpp
     * or agent will block in a endless loop.
     */

    #ifdef WARNING
    fprintf(stderr, "nxagentValidateGC: WARNING! Virtual drawable at [%p] has invalid bits per pixel.\n",
                (void *) pVirtual);

    fprintf(stderr, "nxagentValidateGC: WARNING! While validating GC at [%p] for drawable at [%p] with changes [%lx].\n",
                (void *) pGC, (void *) pDrawable, changes);

    fprintf(stderr, "nxagentValidateGC: WARNING! Bad drawable at [%p] has type [%s] virtual [%p] bits per pixel [%d].\n",
                (void *) pDrawable, (pDrawable -> type == DRAWABLE_PIXMAP) ? "PIXMAP" : "WINDOW",
                    (void *) pVirtual, pVirtual -> bitsPerPixel);
    #endif
  }
  else
  {
    fbValidateGC(pGC, changes, pVirtual);
  }

  if (pGC->tile.pixmap != lastTile)
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentValidateGC: WARNING! Transforming pixmap at [%p] virtual at [%p] "
                "in virtual pixmap.\n", (void *) nxagentPixmapPriv(pGC -> tile.pixmap) -> pRealPixmap,
                    (void *) nxagentPixmapPriv(pGC -> tile.pixmap) -> pRealPixmap);
    #endif

    #ifdef NXAGENT_MVFB_DEBUG
    fprintf(stderr, "nxagentValidateGC: GC [%lx] new tile [%x] from fb set as virtual\n", pGC, pGC->tile.pixmap);
    #endif

    nxagentPixmapIsVirtual(pGC->tile.pixmap) = True;
    nxagentRealPixmap(pGC->tile.pixmap) = nxagentRealPixmap(lastTile);

    if (nxagentRealPixmap(lastTile))
    {
      nxagentPixmapPriv(nxagentRealPixmap(lastTile))->pVirtualPixmap = pGC->tile.pixmap;
    }
  }

  pGC->stipple = lastStipple;

#endif
}

void nxagentChangeGC(GCPtr pGC, unsigned long mask)
{
  XGCValues values;

#ifdef NXAGENT_GCOPTIMIZE
  int changeFlag=0;
#endif

  if (mask & GCFunction)
  {
    values.function = pGC->alu;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag |= nxagentTestGC(alu, function);
#endif
  }

  if (mask & GCPlaneMask)
  {
    values.plane_mask = pGC->planemask;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(planemask, plane_mask);
#endif
  }

  if (mask & GCForeground)
  {
    values.foreground = nxagentPixel(pGC->fgPixel);
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += (values.foreground == nxagentGCPriv(pGC) -> lastServerValues.foreground)?0:1;
    nxagentGCPriv(pGC) -> lastServerValues.foreground=values.foreground;
#endif
  }

  if (mask & GCBackground)
  {
    values.background = nxagentPixel(pGC->bgPixel);
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += (values.background == nxagentGCPriv(pGC) -> lastServerValues.background)?0:1;
    nxagentGCPriv(pGC) -> lastServerValues.background=values.background;
#endif
  }

  if (mask & GCLineWidth)
  {
    values.line_width = pGC->lineWidth;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(lineWidth, line_width);
#endif
  }

  if (mask & GCLineStyle)
  {
    values.line_style = pGC->lineStyle;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(lineStyle, line_style);
#endif
  }

  if (mask & GCCapStyle)
  {
    values.cap_style = pGC->capStyle;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(capStyle, cap_style);
#endif
  }

  if (mask & GCJoinStyle)
  {
    values.join_style = pGC->joinStyle;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(joinStyle, join_style);
#endif
  }

  if (mask & GCFillStyle)
  {
    values.fill_style = pGC->fillStyle;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(fillStyle, fill_style);
#endif
  }

  if (mask & GCFillRule)
  {
    values.fill_rule = pGC->fillRule;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(fillRule, fill_rule);
#endif
  }

  if (mask & GCTile)
  {
    if (pGC->tileIsPixel)
    {
      mask &= ~GCTile;
    }
    else
    {
      values.tile = nxagentPixmap(pGC->tile.pixmap);

#ifdef NXAGENT_MVFB
      pGC->tile.pixmap = nxagentVirtualPixmap(pGC->tile.pixmap);
#ifdef NXAGENT_MVFB_DEBUG
      fprintf(stderr, "nxagentChangeGC: new tile on GC [%x] tile is [%lx]\n", pGC, pGC->tile.pixmap);
#endif
#endif
    }
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += (values.tile == nxagentGCPriv(pGC) -> lastServerValues.tile)?0:1;
    nxagentGCPriv(pGC) -> lastServerValues.tile = values.tile;
#endif
  }

  if (mask & GCStipple)
  {
    values.stipple = nxagentPixmap(pGC->stipple);

#ifdef NXAGENT_MVFB
    pGC->stipple = nxagentVirtualPixmap(pGC->stipple);
#ifdef NXAGENT_MVFB_DEBUG
      fprintf(stderr, "nxagentChangeGC: new stipple on GC [%x] tile is [%lx]\n", pGC, pGC->stipple);
#endif
#endif
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += (values.stipple == nxagentGCPriv(pGC) -> lastServerValues.stipple)? 0 : 1;
    nxagentGCPriv(pGC) -> lastServerValues.stipple = values.stipple;
#endif
  }

  if (mask & GCTileStipXOrigin)
  {
    values.ts_x_origin = pGC->patOrg.x;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(patOrg.x, ts_x_origin);
#endif
  }

  if (mask & GCTileStipYOrigin)
  {
    values.ts_y_origin = pGC->patOrg.y;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(patOrg.y, ts_y_origin);
#endif
  }

  if (mask & GCFont)
  {
    if (!nxagentFontStruct(pGC -> font))
    {
      mask &= ~GCFont;
    }
    else
    {
      values.font = nxagentFont(pGC->font);
#ifdef NXAGENT_GCOPTIMIZE
      changeFlag += (values.font == nxagentGCPriv(pGC) -> lastServerValues.font)? 0 : 1;
      nxagentGCPriv(pGC) -> lastServerValues.font = values.font;
#endif
    }
  } 

  if (mask & GCSubwindowMode)
  {
    values.subwindow_mode = pGC->subWindowMode;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(subWindowMode, subwindow_mode);
#endif
  }

  if (mask & GCGraphicsExposures)
  {
    values.graphics_exposures = pGC->graphicsExposures;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(graphicsExposures, graphics_exposures);
#endif
  }

  if (mask & GCClipXOrigin)
  {
    values.clip_x_origin = pGC->clipOrg.x;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(clipOrg.x, clip_x_origin);
#endif
  }

  if (mask & GCClipYOrigin)
  {
    values.clip_y_origin = pGC->clipOrg.y;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(clipOrg.y, clip_y_origin);
#endif
  }

  if (mask & GCClipMask)
  { /* this is handled in change clip */
    mask &= ~GCClipMask;
  }

  if (mask & GCDashOffset)
  {
    values.dash_offset = pGC->dashOffset;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(dashOffset, dash_offset);
#endif
  }

  if (mask & GCDashList) {
    mask &= ~GCDashList;
    XSetDashes(nxagentDisplay, nxagentGC(pGC),
               pGC->dashOffset, (char *)pGC->dash, pGC->numInDashList);
  }

  if (mask & GCArcMode)
  {
    values.arc_mode = pGC->arcMode;
#ifdef NXAGENT_GCOPTIMIZE
    changeFlag += nxagentTestGC(arcMode, arc_mode);
#endif
  }

#ifdef NXAGENT_GCOPTIMIZE
  if (mask && changeFlag)
  {
    XChangeGC(nxagentDisplay, nxagentGC(pGC), mask, &values);
  }
#ifdef NXAGENT_GCOPTIMIZE_DEBUG
  else if (mask)
    fprintf(stderr, "ChangeGC: discarded = %d Mask=%d\n", ++nDiscarded, mask);
#endif
#else /*GCOPTIMIZE*/
  if (mask)
  {
    XChangeGC(nxagentDisplay, nxagentGC(pGC), mask, &values);
  }

#endif
}

void nxagentCopyGC(GCPtr pGCSrc, unsigned long mask, GCPtr pGCDst)
{
#ifdef NXAGENT_MVFB
#ifdef NXAGENT_MVFB_DEBUG
   fprintf(stderr, "nxagentCopyGC: going to copy GC\n");
#endif
/*   miCopyGC(pGCSrc, mask, pGCDst);*/
#endif
  XCopyGC(nxagentDisplay, nxagentGC(pGCSrc), mask, nxagentGC(pGCDst));

#ifdef NXAGENT_GCOPTIMIZE

  /*
   * Copy the field "foreground" private of the GC if GCForeground
   * is set.
   */

  nxagentCopyGCPriv(GCForeground,foreground,pGCSrc,mask,pGCDst);

#endif
}

void nxagentDestroyGC(GCPtr pGC)
{
#ifdef NXAGENT_MVFB_DEBUG
  fprintf(stderr, "nxagentDestroyGC: GC [%lx]\n", pGC);
#endif

  nxagentDeleteGCSafeRec(pGC);
  XFreeGC(nxagentDisplay, nxagentGC(pGC));
  miDestroyGC(pGC);
}

void nxagentChangeClip(GCPtr pGC, int type, pointer pValue, int nRects)
{
  int i, size;
  BoxPtr pBox;
  XRectangle *pRects;

#ifdef NXAGENT_MVFB_DEBUG
   fprintf(stderr, "nxagentChangeClip: going to change clip on GC [%lx]\n", pGC);
#endif

  nxagentDestroyClipHelper(pGC);

  switch(type)
    {
    case CT_NONE:
      XSetClipMask(nxagentDisplay, nxagentGC(pGC), None);
      break;

    case CT_REGION:
      nRects = REGION_NUM_RECTS((RegionPtr)pValue);
      size = nRects * sizeof(*pRects);
      pRects = (XRectangle *) xalloc(size);
      pBox = REGION_RECTS((RegionPtr)pValue);
      for (i = nRects; i-- > 0;) {
        pRects[i].x = pBox[i].x1;
        pRects[i].y = pBox[i].y1;
        pRects[i].width = pBox[i].x2 - pBox[i].x1;
        pRects[i].height = pBox[i].y2 - pBox[i].y1;
      }
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC), 0, 0,
                         pRects, nRects, Unsorted);
      xfree((char *) pRects);
      break;

    case CT_PIXMAP:
      XSetClipMask(nxagentDisplay, nxagentGC(pGC),
                   nxagentPixmap((PixmapPtr)pValue));
      /*
       * Need to change into region, so subsequent uses are with
       * current pixmap contents.
       */

#if defined(NXAGENT_GCHACK) && !defined(NXAGENT_MVFB)
        {
          register RegionPtr pReg;

          pReg = REGION_CREATE(pValue->drawable.pScreen, NULL, 1);

          pGC->clientClip = pReg;

        }
#else
        {
          pGC->clientClip =
            (pointer) (*pGC->pScreen->BitmapToRegion)((PixmapPtr)pValue);
        }
#endif

      (*pGC->pScreen->DestroyPixmap)((PixmapPtr)pValue);
      pValue = pGC->clientClip;
      type = CT_REGION;

      break;

    case CT_UNSORTED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, Unsorted);
      break;

    case CT_YSORTED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, YSorted);
      break;

    case CT_YXSORTED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, YXSorted);
      break;

    case CT_YXBANDED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, YXBanded);
      break;
    }

  switch(type)
    {
    default:
      break;

    case CT_UNSORTED:
    case CT_YSORTED:
    case CT_YXSORTED:
    case CT_YXBANDED:

      /*
       * other parts of server can only deal with CT_NONE,
       * CT_PIXMAP and CT_REGION client clips.
       */

      pGC->clientClip = (pointer) RECTS_TO_REGION(pGC->pScreen, nRects,
                                                  (xRectangle *)pValue, type);
      xfree(pValue);
      pValue = pGC->clientClip;
      type = CT_REGION;

      break;
    }

  pGC->clientClipType = type;
  pGC->clientClip = pValue;
/*
#ifdef NXAGENT_MVFB
   fprintf(stderr, "nxagentChangeClip: going to change clip GC\n");
  if (type == CT_PIXMAP)
    miChangeClip(pGC, type,(type == CT_PIXMAP)?nxagentVirtualPixmap((PixmapPtr)pValue) : pValue, nRects);
#endif
*/
 nxagentGCPriv(pGC)->nClipRects = nRects;
}

void nxagentDestroyClip(GCPtr pGC)
{
#ifdef NXAGENT_MVFB

  miDestroyClip(pGC);

#endif

  if (pGC->clientClipType == CT_PIXMAP)
  {
     (*pGC->pScreen->DestroyPixmap) ((PixmapPtr) (pGC->clientClip));
  }

  nxagentDestroyClipHelper(pGC);
  XSetClipMask(nxagentDisplay, nxagentGC(pGC), None);

  pGC->clientClipType = CT_NONE;
  pGC->clientClip = NULL;
  nxagentGCPriv(pGC)->nClipRects = 0;
}

void nxagentDestroyClipHelper(GCPtr pGC)
{
  switch (pGC->clientClipType)
    {
    default:
    case CT_NONE:
       break;
    case CT_REGION:
      REGION_DESTROY(pGC->pScreen, pGC->clientClip);
      break;
    }
}

void nxagentCopyClip(GCPtr pGCDst, GCPtr pGCSrc)
{
  RegionPtr pRgn;

  #ifdef TEST
  fprintf(stderr, "nxagentCopyClip: Going to copy clip from GC [%p] to GC [%p]\n",
              (void *) pGCDst, (void *) pGCSrc);
  #endif

/*
#ifdef NXAGENT_MVFB
  if (pGCSrc->clientClipType == CT_PIXMAP && pGCDst->clientClipType == CT_PIXMAP);
    miCopyClip(pGCDst, pGCSrc);
#endif
*/

  switch (pGCSrc->clientClipType)
    {
    default:
    case CT_PIXMAP:

      #ifdef WARNING
      fprintf(stderr, "nxagentCopyClip: WARNING! Not incrementing counter for virtual pixmap at [%p].\n",
                  (void *) nxagentVirtualPixmap((PixmapPtr) pGCSrc->clientClip));
      #endif

      /*
       * Note that both the default and the CT_PIXMAP
       * case statements do not have a break, so we end
       * up calling nxagentDestroyClip() as in CT_NONE.
       */

    case CT_NONE:
      nxagentDestroyClip(pGCDst);
      break;

    case CT_REGION:
      pRgn = REGION_CREATE(pGCDst->pScreen, NULL, 1);
      REGION_COPY(pGCDst->pScreen, pRgn, pGCSrc->clientClip);
      nxagentChangeClip(pGCDst, CT_REGION, pRgn, 0);
      break;
    }

}

static struct nxagentGCRec *nxagentGetFirstGC()
{
  struct nxagentGCRec *tmp = nxagentGCList.first;

  if (nxagentGCList.size)
  {
    nxagentGCList.first = nxagentGCList.first -> next;
    nxagentGCList.size--;

    if (nxagentGCList.size == 0)
    {
      nxagentGCList.last = NULL;
    }
  }

  return tmp;
}

static void nxagentFreeGCRec(struct nxagentGCRec *t)
{
  #ifdef TEST
  fprintf(stderr, "nxagentFreeGCRec: Freeing record at %p GC freed at %p.\n",
              (void*)t, (void*)t -> gc);
  #endif

  xfree(t -> gc);
  free(t);
}

static void nxagentRestoreGCRec(struct nxagentGCRec *t)
{
  #ifdef TEST
  fprintf(stderr, "nxagentRestoreGCRec: Freeing record at %p GC freed at %p.\n",
              (void*)t, (void*)t -> gc);
  #endif

  if (nxagentGC(t -> pGC))
  {
    xfree(nxagentGC(t -> pGC));
  }

  nxagentGC(t -> pGC) = t -> gc;
  free(t);
}

static void nxagentAddGCToList(GCPtr pGC)
{
  struct nxagentGCRec *tmp = malloc(sizeof(struct nxagentGCRec));

  if (!tmp)
  {
    FatalError("nxagentAddGCToList: malloc failed.");
  }

  #ifdef TEST
  fprintf(stderr, "nxagentAddGCToList: Adding GC %p to list at memory %p list size is %d.\n",
              (void*)pGC, (void*)tmp, nxagentGCList.size);
  #endif

  tmp -> pGC = pGC;
  tmp -> gc = nxagentGC(pGC);
  tmp -> next = NULL;

  if (nxagentGCList.size == 0 || nxagentGCList.first == NULL || nxagentGCList.last == NULL)
  {
    nxagentGCList.first = tmp;
  }
  else
  {
    nxagentGCList.last -> next = tmp;
  }

  nxagentGCList.last = tmp;
  nxagentGCList.size++;
}

void nxagentFreeGCList()
{
  struct nxagentGCRec *tmp;

  #ifdef TEST
  fprintf(stderr, "nxagentFreeGCList: List size is %d first elt at %p last elt at %p.\n",
              nxagentGCList.size, (void*)nxagentGCList.first, (void*)nxagentGCList.last);
  #endif

  while ((tmp = nxagentGetFirstGC()))
  {
    nxagentFreeGCRec(tmp);
  }
}

static void nxagentRestoreGCList()
{
  struct nxagentGCRec *tmp;

  #ifdef TEST
  fprintf(stderr, "nxagentRestoreGCList: List size is %d first elt at %p last elt at %p.\n",
              nxagentGCList.size, (void*)nxagentGCList.first, (void*)nxagentGCList.last);
  #endif

  while ((tmp = nxagentGetFirstGC()))
  {
    nxagentRestoreGCRec(tmp);
  }
}

static void nxagentGCReconnect(void *param0, XID param1, pointer param2)
{
  XGCValues values;
  unsigned long valuemask;
  GCPtr pGC = (GCPtr) param0;
  Bool *pBool = (Bool*)param2;

  if (pGC == NULL || !*pBool)
    return;

  if (nxagentGC(pGC))
  {
    nxagentAddGCToList(pGC);
  }
  else
  {
    #ifdef WARNING
    fprintf(stderr, "nxagentGCReconnect: GCRec %p doesn't have a valid pointer to GC data.\n", (void*)pGC);
    #endif
  }

  #ifdef NXAGENT_RECONNECT_GC_DEBUG
  fprintf(stderr, "nxagentGCReconnect: %p\n", pGC);
  #endif

  valuemask = 0;
  memset(&values,0,sizeof(XGCValues));
  values.function = pGC->alu;
  valuemask |= GCFunction;
  values.plane_mask = pGC->planemask;
  valuemask |= GCPlaneMask;
  values.foreground = nxagentPixel(pGC->fgPixel);
  valuemask |= GCForeground;
  values.background = nxagentPixel(pGC->bgPixel);
  valuemask |= GCBackground;

  values.line_width = pGC->lineWidth;
  valuemask |= GCLineWidth;
  values.line_style = pGC->lineStyle;
  valuemask |= GCLineStyle;
  values.cap_style = pGC->capStyle;
  valuemask |= GCCapStyle;
  values.join_style = pGC->joinStyle;
  valuemask |= GCJoinStyle;
  values.fill_style = pGC->fillStyle;
  valuemask |= GCFillStyle;
  values.fill_rule = pGC->fillRule;
  valuemask |= GCFillRule;
  if (!pGC->tileIsPixel  && (pGC->tile.pixmap != NULL))
  {
      if (nxagentPixmapIsVirtual(pGC->tile.pixmap))
      values.tile = nxagentPixmap(nxagentRealPixmap(pGC->tile.pixmap));
     else
      values.tile = nxagentPixmap(pGC->tile.pixmap);

      valuemask |= GCTile;
  }
  if (pGC->stipple != NULL)
  {
    
    if (nxagentPixmapIsVirtual(pGC->stipple))
    {
      if (nxagentPixmap(nxagentRealPixmap(pGC -> stipple)) == 0)
        nxagentReconnectPixmap(nxagentRealPixmap(pGC -> stipple), 0, pBool);
      values.stipple = nxagentPixmap(nxagentRealPixmap(pGC->stipple));
    }
    else
    {
      if (nxagentPixmap(pGC -> stipple) == 0)
        nxagentReconnectPixmap(pGC -> stipple, 0, pBool);
      values.stipple = nxagentPixmap(pGC->stipple);
    }
    valuemask |= GCStipple;
  }
  values.ts_x_origin = pGC->patOrg.x;
  valuemask |= GCTileStipXOrigin;
  values.ts_y_origin = pGC->patOrg.y;
  valuemask |= GCTileStipYOrigin;
  if (pGC->font != NULL)
  {
    values.font = nxagentFont(pGC->font);
    valuemask |= GCFont;
  }
  values.subwindow_mode = pGC->subWindowMode;
  valuemask |= GCSubwindowMode;
  values.graphics_exposures = pGC->graphicsExposures;
  valuemask |= GCGraphicsExposures;
  values.clip_x_origin = pGC->clipOrg.x;
  valuemask |= GCClipXOrigin;
  values.clip_y_origin = pGC->clipOrg.y;
  valuemask |= GCClipYOrigin;
  valuemask |= GCClipMask;
  values.dash_offset = pGC->dashOffset;
  valuemask |= GCDashOffset;
  if (pGC->dash != NULL)
  {
     values.dashes = *pGC->dash;
     valuemask |= GCDashList;
  }
  values.arc_mode = pGC->arcMode;
  valuemask |= GCArcMode;

  if ((nxagentGC(pGC) = XCreateGC(nxagentDisplay,
                                nxagentDefaultDrawables[pGC->depth],
                                valuemask, &values)) == NULL)
    *pBool = False;

  nxagentDummyChangeClip(pGC, 
                    pGC -> clientClipType,
                    pGC -> clientClip,
                    nxagentGCPriv(pGC) -> nClipRects);

  #ifdef NXAGENT_RECONNECT_DEBUG
  XSync(nxagentDisplay, False);
  #endif
}

Bool nxagentReconnectAllGCs(void *p0)
{
  int flexibility;
  int cid;
  Bool GCSuccess = True;

  flexibility = *(int*)p0;

  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_GC_DEBUG)
  fprintf(stderr, "nxagentReconnectAllGCs\n");
  #endif

  for (cid = 0; cid < MAXCLIENTS; cid++)
  {
    if (clients[cid] && GCSuccess)
    {
      FindClientResourcesByType(clients[cid], RT_GC, nxagentGCReconnect, &GCSuccess);
    }
  }

  nxagentGCSafeVectorReconnect();

  return GCSuccess;
}

void nxagentDisconnectGC(pointer p0, XID x1, pointer p2)
{
  GCPtr pGC = (GCPtr) p0;
  Bool* pBool = (Bool*) p2;

  if (!*pBool || !pGC)
  {
    if (!pGC) 
    {
      fprintf(stderr, "nxagentDisconnectGC: pGC is NULL\n");
    }

    return;
  }

  if (pGC -> stipple)
  {
    PixmapPtr pMap = pGC -> stipple;

    nxagentDisconnectPixmap(nxagentRealPixmap(pMap), 0, pBool);
  }
} 

Bool nxagentDisconnectAllGCs()
{
  int cid;
  Bool success = True;

  #if defined(NXAGENT_RECONNECT_DEBUG) || defined(NXAGENT_RECONNECT_GC_DEBUG)
  fprintf(stderr, "nxagentDisconnectAllGCs\n");
  #endif

  for (cid = 0; (cid < MAXCLIENTS) && success; cid++)
  {
    if (clients[cid])
    {
      FindClientResourcesByType(clients[cid], RT_GC,
                                    (FindResType) nxagentDisconnectGC, &success);
    }
  }

  nxagentRestoreGCList();

  return success;
}

Bool nxagentScratchGCsDestroy()
{
  int i;
  ScreenPtr pScreen = screenInfo.screens[0];
  Bool succeding = True;

  for (i = 0; (i <= pScreen->numDepths) && succeding; i++)
  {
    #ifdef NXAGENT_RECONNECT_GC_DEBUG
    fprintf(stderr, "nxagentScratchGCsDestroy: Destroying scratch for depth %d\n",
                pScreen->GCperDepth[i]->depth);
    #endif

    nxagentDisconnectGC(pScreen->GCperDepth[i], (XID)0, &succeding);
  }

  return succeding;
}

Bool nxagentScratchGCsCreate()
{
  int i;
  ScreenPtr pScreen = screenInfo.screens[0];
  Bool succeding = True;

  for (i = 0; (i <= pScreen->numDepths) && succeding; i++)
  {
    #ifdef NXAGENT_RECONNECT_GC_DEBUG
    fprintf(stderr, "nxagentScratchGCsCreate: Reconnecting scratch gc for depth %d used %d\n",
             pScreen->GCperDepth[i]->depth, (pScreen->rgf & (1L << (i+1))));
    #endif

    nxagentGCReconnect(pScreen->GCperDepth[i], (XID)0, &succeding);
    nxagentDeleteGCSafeRec(pScreen->GCperDepth[i]);

    /*
     * Shall i have any problem
     * if i reconnect a not used scratch GC ?
     *
     * if ((pScreen->rgf & (1L << (i+1))))
     * {
     *   nxagentGCReconnect(pScreen->GCperDepth[i], (XID)0, &succeding);
     * }
     * else if(!(nxagentGC(pScreen->GCperDepth[i]) = 
     *             XCreateGC(nxagentDisplay,
     *                           nxagentDefaultDrawables[pScreen->GCperDepth[i]->depth],
     *                               0, NULL)))
     * {
     *   succeding = False;
     * }
     */
  }

  return succeding;
}

static void nxagentDummyChangeClip(GCPtr pGC, int type, pointer pValue, int nRects)
{
  int i, size;
  BoxPtr pBox;
  XRectangle *pRects;

  #ifdef NXAGENT_MVFB_DEBUG
  fprintf(stderr, "nxagentChangeClip: going to change clip on GC [%lx]\n", pGC);
  #endif

  /* nxagentDestroyClipHelper(pGC); */
  #ifdef NXAGENT_RECONNECT_GC_DEBUG
  fprintf(stderr, "nxagentDummyChangeClip: type is %s\n",
              (type == CT_NONE) ? "CT_NONE" :
                  (type == CT_REGION) ? "CT_REGION" :
                      (type == CT_PIXMAP) ? "CT_REGION" : "UNKNOWN");
  #endif

  switch(type)
    {
    case CT_NONE:
      XSetClipMask(nxagentDisplay, nxagentGC(pGC), None);
      break;

    case CT_REGION:
      nRects = REGION_NUM_RECTS((RegionPtr)pValue);
      size = nRects * sizeof(*pRects);
      pRects = (XRectangle *) xalloc(size);
      pBox = REGION_RECTS((RegionPtr)pValue);
      for (i = nRects; i-- > 0;) {
        pRects[i].x = pBox[i].x1;
        pRects[i].y = pBox[i].y1;
        pRects[i].width = pBox[i].x2 - pBox[i].x1;
        pRects[i].height = pBox[i].y2 - pBox[i].y1;
      }
      /* FIXME: originally the clip origin area were 0,0 but it didn't work with
       * kedit and family, because it got the clip mask of the pixmap all traslated. */
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC), pGC -> clipOrg.x, pGC -> clipOrg.y,
                         pRects, nRects, Unsorted);
      xfree((char *) pRects);
      break;

    case CT_PIXMAP:
      XSetClipMask(nxagentDisplay, nxagentGC(pGC),
                   nxagentPixmap((PixmapPtr)pValue));
      /*
       * Need to change into region, so subsequent uses are with
       * current pixmap contents.
       */

#if defined(NXAGENT_GCHACK) && !defined(NXAGENT_MVFB)
        {
          register RegionPtr pReg;

          pReg = REGION_CREATE(pValue->drawable.pScreen, NULL, 1);

          pGC->clientClip = pReg;

        }
#else
        {
          pGC->clientClip =
            (pointer) (*pGC->pScreen->BitmapToRegion)((PixmapPtr)pValue);
        }
#endif

      (*pGC->pScreen->DestroyPixmap)((PixmapPtr)pValue);
      pValue = pGC->clientClip;
      type = CT_REGION;

      break;

    case CT_UNSORTED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, Unsorted);
      break;

    case CT_YSORTED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, YSorted);
      break;

    case CT_YXSORTED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, YXSorted);
      break;

    case CT_YXBANDED:
      XSetClipRectangles(nxagentDisplay, nxagentGC(pGC),
                         pGC->clipOrg.x, pGC->clipOrg.y,
                         (XRectangle *)pValue, nRects, YXBanded);
      break;
    }

  switch(type)
    {
    default:
      break;

    case CT_UNSORTED:
    case CT_YSORTED:
    case CT_YXSORTED:
    case CT_YXBANDED:

      /*
       * other parts of server can only deal with CT_NONE,
       * CT_PIXMAP and CT_REGION client clips.
       */

      pGC->clientClip = (pointer) RECTS_TO_REGION(pGC->pScreen, nRects,
                                                  (xRectangle *)pValue, type);
      xfree(pValue);
      pValue = pGC->clientClip;
      type = CT_REGION;

      break;
    }

  pGC->clientClipType = type;
  pGC->clientClip = pValue;
/*
#ifdef NXAGENT_MVFB
   fprintf(stderr, "nxagentChangeClip: going to change clip GC\n");
  if (type == CT_PIXMAP)
    miChangeClip(pGC, type,(type == CT_PIXMAP)?nxagentVirtualPixmap((PixmapPtr)pValue) : pValue, nRects);
#endif
*/
 nxagentGCPriv(pGC)->nClipRects = nRects;
}
